# type: ignore  # Zarr Python v2 types are different to v3
from typing import Optional, Union

import zarr
from numcodecs.registry import get_codec

from cubed.types import T_DType, T_RegularChunks, T_Shape, T_Store
from cubed.utils import join_path


class ZarrArrayGroup(dict):
    def __init__(
        self,
        shape: Optional[T_Shape] = None,
        dtype: Optional[T_DType] = None,
        chunks: Optional[T_RegularChunks] = None,
    ):
        dict.__init__(self)
        self.shape = shape
        self.dtype = dtype
        self.chunks = chunks

    def __getitem__(self, key):
        if isinstance(key, str):
            return super().__getitem__(key)
        return {field: zarray[key] for field, zarray in self.items()}

    def set_basic_selection(self, selection, value, fields=None):
        self[fields][selection] = value


def open_zarr_array(
    store: T_Store,
    mode: str,
    *,
    shape: Optional[T_Shape] = None,
    dtype: Optional[T_DType] = None,
    chunks: Optional[T_RegularChunks] = None,
    path: Optional[str] = None,
    compressor: Union[dict, str, None] = "default",
    **kwargs,
):
    if isinstance(compressor, dict):
        compressor = get_codec(compressor)

    if dtype is None or not hasattr(dtype, "fields") or dtype.fields is None:
        return zarr.open_array(
            store,
            mode=mode,
            shape=shape,
            dtype=dtype,
            chunks=chunks,
            path=path,
            compressor=compressor,
            **kwargs,
        )
    else:
        ret = ZarrArrayGroup(shape=shape, dtype=dtype, chunks=chunks)
        for field in dtype.fields:
            field_dtype, _ = dtype.fields[field]
            field_path = field if path is None else join_path(path, field)
            ret[field] = zarr.open_array(
                store,
                mode=mode,
                shape=shape,
                dtype=field_dtype,
                chunks=chunks,
                path=field_path,
                compressor=compressor,
                **kwargs,
            )
        return ret



from typing import Any, Mapping, Optional, Sequence

import coiled
from networkx import MultiDiGraph

from cubed.runtime.pipeline import visit_nodes
from cubed.runtime.types import Callback, DagExecutor
from cubed.runtime.utils import (
    execution_stats,
    handle_callbacks,
    handle_operation_start_callbacks,
)
from cubed.spec import Spec


def make_coiled_function(func, coiled_kwargs):
    return coiled.function(**coiled_kwargs)(execution_stats(func))


class CoiledExecutor(DagExecutor):
    """An execution engine that uses Coiled Functions."""

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    @property
    def name(self) -> str:
        return "coiled"

    def execute_dag(
        self,
        dag: MultiDiGraph,
        callbacks: Optional[Sequence[Callback]] = None,
        spec: Optional[Spec] = None,
        compute_id: Optional[str] = None,
        **coiled_kwargs: Mapping[str, Any],
    ) -> None:
        merged_kwargs = {**self.kwargs, **coiled_kwargs}
        minimum_workers = merged_kwargs.pop("minimum_workers", None)
        # Note this currently only builds the task graph for each stage once it gets to that stage in computation
        for name, node in visit_nodes(dag):
            handle_operation_start_callbacks(callbacks, name)
            pipeline = node["pipeline"]
            coiled_function = make_coiled_function(pipeline.function, merged_kwargs)
            if minimum_workers is not None:
                coiled_function.cluster.adapt(minimum=minimum_workers)
            # coiled expects a sequence (it calls `len` on it)
            input = list(pipeline.mappable)
            for result, stats in coiled_function.map(input, config=pipeline.config):
                if callbacks is not None:
                    if name is not None:
                        stats["name"] = name
                    handle_callbacks(callbacks, result, stats)



import atexit
import dataclasses
import inspect
import shutil
import tempfile
import uuid
from datetime import datetime
from functools import lru_cache
from typing import Any, Callable, Dict, Optional, Tuple

import networkx as nx

from cubed.core.optimization import multiple_inputs_optimize_dag
from cubed.primitive.blockwise import BlockwiseSpec
from cubed.primitive.types import PrimitiveOperation
from cubed.runtime.pipeline import visit_nodes
from cubed.runtime.types import ComputeEndEvent, ComputeStartEvent, CubedPipeline
from cubed.storage.backend import is_backend_storage_array
from cubed.storage.zarr import LazyZarrArray, open_if_lazy_zarr_array
from cubed.utils import (
    chunk_memory,
    extract_stack_summaries,
    is_local_path,
    itemsize,
    join_path,
    memory_repr,
)

try:
    from zarr.errors import ArrayNotFoundError  # type: ignore
except ImportError:
    ArrayNotFoundError = FileNotFoundError  # zarr-python 3

# A unique ID with sensible ordering, used for making directory names
CONTEXT_ID = f"cubed-{datetime.now().strftime('%Y%m%dT%H%M%S')}-{uuid.uuid4()}"

# Delete local context dirs when Python exits
CONTEXT_DIRS = set()

Decorator = Callable


def delete_on_exit(context_dir: str) -> None:
    if context_dir not in CONTEXT_DIRS and is_local_path(context_dir):
        atexit.register(lambda: shutil.rmtree(context_dir, ignore_errors=True))
        CONTEXT_DIRS.add(context_dir)


sym_counter = 0


def gensym(name="op"):
    global sym_counter
    sym_counter += 1
    return f"{name}-{sym_counter:03}"


class Plan:
    """Deferred computation plan for a graph of arrays.

    A thin wrapper around a NetworkX `MultiDiGraph`. Nodes are Zarr paths, and may
    have a `pipeline` attribute holding the pipeline to execute to generate
    the output at that path. If there is no `pipeline` attribute no computation
    is needed since the Zarr file already exists. Directed edges point towards
    the dependent Zarr path.

    Multiple edges are possible since a node (Zarr file) may be computed by
    a function with repeated inputs. For example, consider `equals` where the
    two arguments are the same array. We need to keep track of these cases, so
    we use a NetworkX `MultiDiGraph` rather than just as `DiGraph`.

    Compared to a more traditional DAG representing a computation, in Cubed
    nodes are not values that are passed to functions, they are instead
    "parallel computations" which are run for their side effects. Data does
    not flow through the graph - it is written to external storage (Zarr files)
    as the output of one pipeline, then read back as the input to later pipelines.
    """

    def __init__(self, dag):
        self.dag = dag

    # args from primitive_op onwards are omitted for creation functions when no computation is needed
    @classmethod
    def _new(
        cls,
        name,
        op_name,
        target,
        primitive_op=None,
        hidden=False,
        *source_arrays,
    ):
        # create an empty DAG or combine from sources
        if len(source_arrays) == 0:
            dag = nx.MultiDiGraph()
        else:
            dag = arrays_to_dag(*source_arrays)

        # add new node and edges
        frame = inspect.currentframe().f_back  # go back one in the stack
        stack_summaries = extract_stack_summaries(frame, limit=10)

        first_cubed_i = min(i for i, s in enumerate(stack_summaries) if s.is_cubed())
        first_cubed_summary = stack_summaries[first_cubed_i]

        op_name_unique = gensym()

        if primitive_op is None:
            # op
            dag.add_node(
                op_name_unique,
                name=op_name_unique,
                op_name=op_name,
                type="op",
                stack_summaries=stack_summaries,
                op_display_name=f"{op_name_unique}\n{first_cubed_summary.name}",
                hidden=hidden,
            )
            # array
            if isinstance(name, list):  # multiple outputs
                for n, t in zip(name, target):
                    dag.add_node(
                        n,
                        name=n,
                        type="array",
                        target=t,
                        hidden=hidden,
                    )
                    dag.add_edge(op_name_unique, n)
            else:  # single output
                dag.add_node(
                    name,
                    name=name,
                    type="array",
                    target=target,
                    hidden=hidden,
                )
                dag.add_edge(op_name_unique, name)
        else:
            # op
            dag.add_node(
                op_name_unique,
                name=op_name_unique,
                op_name=op_name,
                type="op",
                stack_summaries=stack_summaries,
                op_display_name=f"{op_name_unique}\n{first_cubed_summary.name}",
                hidden=hidden,
                primitive_op=primitive_op,
                pipeline=primitive_op.pipeline,
            )
            # array
            if isinstance(name, list):  # multiple outputs
                for n, t in zip(name, target):
                    dag.add_node(
                        n,
                        name=n,
                        type="array",
                        target=t,
                        hidden=hidden,
                    )
                    dag.add_edge(op_name_unique, n)
            else:  # single output
                dag.add_node(
                    name,
                    name=name,
                    type="array",
                    target=target,
                    hidden=hidden,
                )
                dag.add_edge(op_name_unique, name)
        for x in source_arrays:
            if hasattr(x, "name"):
                dag.add_edge(x.name, op_name_unique)

        return Plan(dag)

    @classmethod
    def arrays_to_plan(cls, *arrays):
        return Plan(arrays_to_dag(*arrays))

    def optimize(
        self,
        optimize_function: Optional[Callable[..., nx.MultiDiGraph]] = None,
        array_names: Optional[Tuple[str]] = None,
    ):
        if optimize_function is None:
            optimize_function = multiple_inputs_optimize_dag
        dag = optimize_function(self.dag, array_names=array_names)
        return Plan(dag)

    def _create_lazy_zarr_arrays(self, dag):
        # find all lazy zarr arrays in dag
        all_pipeline_nodes = []
        lazy_zarr_arrays = []
        allowed_mem = 0
        reserved_mem = 0
        for n, d in dag.nodes(data=True):
            if "primitive_op" in d:
                all_pipeline_nodes.append(n)
                allowed_mem = max(allowed_mem, d["primitive_op"].allowed_mem)
                reserved_mem = max(reserved_mem, d["primitive_op"].reserved_mem)

            if "target" in d and isinstance(d["target"], LazyZarrArray):
                lazy_zarr_arrays.append(d["target"])

        if len(lazy_zarr_arrays) > 0:
            # add new node and edges
            name = "create-arrays"
            op_name = name
            primitive_op = create_zarr_arrays(
                lazy_zarr_arrays, allowed_mem, reserved_mem
            )
            dag.add_node(
                name,
                name=name,
                op_name=op_name,
                type="op",
                op_display_name=name,
                primitive_op=primitive_op,
                pipeline=primitive_op.pipeline,
            )
            dag.add_node(
                "arrays",
                name="arrays",
                target=None,
            )
            dag.add_edge(name, "arrays")
            # make create arrays node a predecessor of all pipeline nodes so it runs first
            for n in all_pipeline_nodes:
                dag.add_edge("arrays", n)

        return dag

    def _compile_blockwise(self, dag, compile_function: Decorator) -> nx.MultiDiGraph:
        """Compiles functions from all blockwise ops by mutating the input dag."""
        # Recommended: make a copy of the dag before calling this function.

        compile_with_config = (
            "config" in inspect.getfullargspec(compile_function).kwonlyargs
        )

        for n in dag.nodes:
            node = dag.nodes[n]

            if "primitive_op" not in node:
                continue

            if not isinstance(node["pipeline"].config, BlockwiseSpec):
                continue

            if compile_with_config:
                compiled = compile_function(
                    node["pipeline"].config.function, config=node["pipeline"].config
                )
            else:
                compiled = compile_function(node["pipeline"].config.function)

            # node is a blockwise primitive_op.
            # maybe we should investigate some sort of optics library for frozen dataclasses...
            new_pipeline = dataclasses.replace(
                node["pipeline"],
                config=dataclasses.replace(node["pipeline"].config, function=compiled),
            )
            node["pipeline"] = new_pipeline

        return dag

    @lru_cache  # noqa: B019
    def _finalize(
        self,
        optimize_graph: bool = True,
        optimize_function=None,
        compile_function: Optional[Decorator] = None,
        array_names=None,
    ) -> "FinalizedPlan":
        dag = (
            self.optimize(optimize_function, array_names).dag
            if optimize_graph
            else self.dag
        )
        # create a copy since _create_lazy_zarr_arrays mutates the dag
        dag = dag.copy()
        if callable(compile_function):
            dag = self._compile_blockwise(dag, compile_function)
        dag = self._create_lazy_zarr_arrays(dag)
        return FinalizedPlan(nx.freeze(dag))

    def execute(
        self,
        executor=None,
        callbacks=None,
        optimize_graph=True,
        optimize_function=None,
        compile_function=None,
        resume=None,
        array_names=None,
        spec=None,
        **kwargs,
    ):
        finalized_plan = self._finalize(
            optimize_graph, optimize_function, compile_function, array_names=array_names
        )
        dag = finalized_plan.dag

        if resume:
            # mark nodes as computed so they are not visited by visit_nodes
            dag = dag.copy()
            nodes = {n: d for (n, d) in dag.nodes(data=True)}
            for name in list(nx.topological_sort(dag)):
                dag.nodes[name]["computed"] = already_computed(name, dag, nodes)

        compute_id = f"compute-{datetime.now().strftime('%Y%m%dT%H%M%S.%f')}"

        if callbacks is not None:
            event = ComputeStartEvent(compute_id, dag)
            [callback.on_compute_start(event) for callback in callbacks]
        executor.execute_dag(
            dag,
            compute_id=compute_id,
            callbacks=callbacks,
            spec=spec,
            **kwargs,
        )
        if callbacks is not None:
            event = ComputeEndEvent(compute_id, dag)
            [callback.on_compute_end(event) for callback in callbacks]

    def visualize(
        self,
        filename="cubed",
        format=None,
        rankdir="TB",
        optimize_graph=True,
        optimize_function=None,
        show_hidden=False,
        array_names=None,
    ):
        finalized_plan = self._finalize(
            optimize_graph, optimize_function, array_names=array_names
        )
        dag = finalized_plan.dag
        dag = dag.copy()  # make a copy since we mutate the DAG below

        # remove edges from create-arrays output node to avoid cluttering the diagram
        dag.remove_edges_from(list(dag.out_edges("arrays")))

        if not show_hidden:
            dag.remove_nodes_from(
                list(n for n, d in dag.nodes(data=True) if d.get("hidden", False))
            )

        dag.graph["graph"] = {
            "rankdir": rankdir,
            "label": (
                # note that \l is used to left-justify each line (see https://www.graphviz.org/docs/attrs/nojustify/)
                rf"num tasks: {finalized_plan.num_tasks()}\l"
                rf"max projected memory: {memory_repr(finalized_plan.max_projected_mem())}\l"
                rf"total nbytes written: {memory_repr(finalized_plan.total_nbytes_written())}\l"
                rf"optimized: {optimize_graph}\l"
            ),
            "labelloc": "bottom",
            "labeljust": "left",
            "fontsize": "10",
        }
        dag.graph["node"] = {"fontname": "helvetica", "shape": "box", "fontsize": "10"}

        # do an initial pass to extract array variable names from stack summaries
        array_display_names = {}
        for _, d in dag.nodes(data=True):
            if "stack_summaries" in d:
                stack_summaries = d["stack_summaries"]
                first_cubed_i = min(
                    i for i, s in enumerate(stack_summaries) if s.is_cubed()
                )
                caller_summary = stack_summaries[first_cubed_i - 1]
                array_display_names.update(caller_summary.array_names_to_variable_names)
        # add current stack info
        frame = inspect.currentframe().f_back  # go back one in the stack
        stack_summaries = extract_stack_summaries(frame, limit=10)
        first_cubed_i = min(i for i, s in enumerate(stack_summaries) if s.is_cubed())
        caller_summary = stack_summaries[first_cubed_i - 1]
        array_display_names.update(caller_summary.array_names_to_variable_names)

        # now set node attributes with visualization info
        for n, d in dag.nodes(data=True):
            label = n
            tooltip = f"name: {n}\n"
            node_type = d.get("type", None)
            if node_type == "op":
                label = d["op_display_name"]
                op_name = d["op_name"]
                if op_name == "blockwise":
                    d["style"] = '"rounded,filled"'
                    d["fillcolor"] = "#dcbeff"
                elif op_name == "rechunk":
                    d["style"] = '"rounded,filled"'
                    d["fillcolor"] = "#aaffc3"
                else:
                    # creation function
                    d["style"] = "rounded"
                tooltip += f"op: {op_name}"

                num_tasks = None
                if "primitive_op" in d:
                    primitive_op = d["primitive_op"]
                    tooltip += (
                        f"\nprojected memory: {memory_repr(primitive_op.projected_mem)}"
                    )
                    num_tasks = primitive_op.num_tasks
                    tooltip += f"\ntasks: {num_tasks}"
                    if primitive_op.write_chunks is not None:
                        tooltip += f"\nwrite chunks: {primitive_op.write_chunks}"
                    del d["primitive_op"]

                # remove pipeline attribute since it is a long string that causes graphviz to fail
                if "pipeline" in d:
                    pipeline = d["pipeline"]
                    if isinstance(pipeline.config, BlockwiseSpec):
                        tooltip += (
                            f"\nnum input blocks: {pipeline.config.num_input_blocks}"
                        )
                        tooltip += (
                            f"\nnum output blocks: {pipeline.config.num_output_blocks}"
                        )
                    del d["pipeline"]

                if "stack_summaries" in d and d["stack_summaries"] is not None:
                    # add call stack information
                    stack_summaries = d["stack_summaries"]

                    first_cubed_i = min(
                        i for i, s in enumerate(stack_summaries) if s.is_cubed()
                    )
                    caller_summary = stack_summaries[first_cubed_i - 1]

                    calls = " -> ".join(
                        [
                            s.name
                            for s in stack_summaries
                            if not s.is_on_python_lib_path()
                        ]
                    )

                    line = f"{caller_summary.lineno} in {caller_summary.name}"

                    tooltip += f"\ncalls: {calls}"
                    tooltip += f"\nline: {line}"
                    del d["stack_summaries"]

                if num_tasks is not None:
                    label += f"\ntasks: {num_tasks}"

            elif node_type == "array":
                target = d["target"]
                chunkmem = memory_repr(chunk_memory(target))

                # materialized arrays are light orange, virtual arrays are white
                if isinstance(target, LazyZarrArray) or is_backend_storage_array(
                    target
                ):
                    d["style"] = "filled"
                    d["fillcolor"] = "#ffd8b1"
                if n in array_display_names:
                    var_name = array_display_names[n]
                    label = f"{n}\n{var_name}"
                    tooltip += f"variable: {var_name}\n"
                tooltip += f"shape: {target.shape}\n"
                tooltip += f"chunks: {target.chunks}\n"
                tooltip += f"dtype: {target.dtype}\n"
                tooltip += f"chunk memory: {chunkmem}\n"
                if hasattr(target, "nbytes"):
                    tooltip += f"nbytes: {memory_repr(target.nbytes)}\n"
                if hasattr(target, "nchunks"):
                    tooltip += f"nchunks: {target.nchunks}\n"

                del d["target"]

            # quote strings with colons in them (https://github.com/pydot/pydot/issues/258)
            d["label"] = '"' + label.strip() + '"'
            d["tooltip"] = '"' + tooltip.strip() + '"'

            if "name" in d:  # pydot already has name
                del d["name"]
        gv = nx.drawing.nx_pydot.to_pydot(dag)
        if format is None:
            format = "svg"
        full_filename = f"{filename}.{format}"
        gv.write(full_filename, format=format)

        try:  # pragma: no cover
            import IPython.display as display

            if format == "svg":
                return display.HTML(filename=full_filename)
        except ImportError:
            # Can't return a display object if no IPython.
            pass
        return None


class FinalizedPlan:
    """A plan that is ready to be run.

    Finalizing a plan involves the following steps:
    1. optimization (optional)
    2. adding housekeping nodes to create arrays
    3. compiling functions (optional)
    4. freezing the final DAG so it can't be changed
    """

    def __init__(self, dag):
        self.dag = dag

    def max_projected_mem(self):
        """Return the maximum projected memory across all tasks to execute this plan."""
        projected_mem_values = [
            node["primitive_op"].projected_mem for _, node in visit_nodes(self.dag)
        ]
        return max(projected_mem_values) if len(projected_mem_values) > 0 else 0

    def num_arrays(self) -> int:
        """Return the number of arrays in this plan."""
        return sum(d.get("type") == "array" for _, d in self.dag.nodes(data=True))

    def num_primitive_ops(self) -> int:
        """Return the number of primitive operations in this plan."""
        return len(list(visit_nodes(self.dag)))

    def num_tasks(self):
        """Return the number of tasks needed to execute this plan."""
        tasks = 0
        for _, node in visit_nodes(self.dag):
            tasks += node["primitive_op"].num_tasks
        return tasks

    def total_nbytes_written(self) -> int:
        """Return the total number of bytes written for all materialized arrays in this plan."""
        nbytes = 0
        for _, d in self.dag.nodes(data=True):
            if d.get("type") == "array":
                target = d["target"]
                if isinstance(target, LazyZarrArray):
                    nbytes += target.nbytes
        return nbytes


def arrays_to_dag(*arrays):
    from .array import check_array_specs

    check_array_specs(arrays)
    dags = [x.plan.dag for x in arrays if hasattr(x, "plan")]
    return nx.compose_all(dags)


def arrays_to_plan(*arrays):
    plans = [x.plan for x in arrays if hasattr(x, "plan")]
    if len(plans) == 0:
        raise ValueError(f"No plans found for arrays: {arrays}")
    return plans[0].arrays_to_plan(*arrays)


def context_dir_path(spec=None):
    work_dir = spec.work_dir if spec is not None else None
    if work_dir is None:
        work_dir = tempfile.gettempdir()
    context_dir = join_path(work_dir, CONTEXT_ID)
    delete_on_exit(context_dir)
    return context_dir


def new_temp_path(name, suffix=".zarr", spec=None):
    """Return a string path for a temporary file path, which may be local or remote.

    Note that this function does not create the file or any directories (and they
    may never be created, if for example the file doesn't need to be materialized).
    """
    context_dir = context_dir_path(spec)
    return join_path(context_dir, f"{name}{suffix}")


def create_zarr_array(lazy_zarr_array, *, config=None):
    """Stage function for create."""
    lazy_zarr_array.create(mode="a")


def create_zarr_arrays(lazy_zarr_arrays, allowed_mem, reserved_mem):
    # projected memory is size of largest dtype size (for a fill value)
    projected_mem = (
        max([itemsize(lza.dtype) for lza in lazy_zarr_arrays], default=0) + reserved_mem
    )
    num_tasks = len(lazy_zarr_arrays)

    pipeline = CubedPipeline(
        create_zarr_array,
        "create_zarr_array",
        lazy_zarr_arrays,
        None,
    )
    return PrimitiveOperation(
        pipeline=pipeline,
        source_array_names=[],
        target_array=None,
        projected_mem=projected_mem,
        allowed_mem=allowed_mem,
        reserved_mem=reserved_mem,
        num_tasks=num_tasks,
        fusable_with_predecessors=False,
    )


def already_computed(name, dag, nodes: Dict[str, Any]) -> bool:
    """
    Return True if the array for a node doesn't have a pipeline to compute it,
    or if all its outputs have already been computed (all chunks are present).
    """
    pipeline = nodes[name].get("pipeline", None)
    if pipeline is None:
        return True

    # if no outputs have targets then need to compute (this is the create-arrays case)
    if all(
        [nodes[output].get("target", None) is None for output in dag.successors(name)]
    ):
        return False

    for output in dag.successors(name):
        target = nodes[output].get("target", None)
        if target is not None:
            try:
                target = open_if_lazy_zarr_array(target)
                if not hasattr(target, "nchunks_initialized"):
                    raise NotImplementedError(
                        f"Zarr array type {type(target)} does not support resume since it doesn't have a 'nchunks_initialized' property"
                    )
                # this check can be expensive since it has to list the directory to find nchunks_initialized
                if target.ndim == 0 or target.nchunks_initialized != target.nchunks:
                    return False
            except ArrayNotFoundError:
                return False
    return True



from cubed.array.overlap import map_overlap
from cubed.array_api.creation_functions import asarray
from cubed.array_api.manipulation_functions import concat
from cubed.backend_array_api import namespace as nxp
from cubed.core import reduction
from cubed.vendor.dask.array.core import normalize_chunks
from cubed.vendor.dask.array.utils import validate_axis


def all(x, /, *, axis=None, keepdims=False, split_every=None):
    if x.size == 0:
        return asarray(True, dtype=x.dtype)
    return reduction(
        x,
        nxp.all,
        axis=axis,
        dtype=nxp.bool,
        keepdims=keepdims,
        split_every=split_every,
    )


def any(x, /, *, axis=None, keepdims=False, split_every=None):
    if x.size == 0:
        return asarray(False, dtype=x.dtype)
    return reduction(
        x,
        nxp.any,
        axis=axis,
        dtype=nxp.bool,
        keepdims=keepdims,
        split_every=split_every,
    )


def diff(x, /, *, axis=-1, n=1, prepend=None, append=None):
    axis = validate_axis(axis, x.ndim)

    if n < 0:
        raise ValueError(f"order of diff must be non-negative, but was {n}")
    if n == 0:
        return x

    combined = []
    if prepend is not None:
        combined.append(prepend)
    combined.append(x)
    if append is not None:
        combined.append(append)
    if len(combined) > 1:
        x = concat(combined, axis=axis, chunks=x.chunksize)

    shape = tuple(s - n if i == axis else s for i, s in enumerate(x.shape))
    chunks = normalize_chunks(x.chunksize, shape, dtype=x.dtype)
    depth = {axis: (0, n)}  # only need look-ahead values for differencing
    return map_overlap(
        nxp.diff,
        x,
        dtype=x.dtype,
        chunks=chunks,
        depth=depth,
        axis=axis,
        n=n,
    )



import pytest
import zarr
from numcodecs.registry import get_codec

from cubed import config
from cubed.storage.backend import open_backend_array
from cubed.storage.zarr import lazy_zarr_array

ZARR_PYTHON_V3 = zarr.__version__[0] == "3"


def test_lazy_zarr_array(tmp_path):
    zarr_path = tmp_path / "lazy.zarr"
    arr = lazy_zarr_array(zarr_path, shape=(3, 3), dtype=int, chunks=(2, 2))

    assert not zarr_path.exists()
    with pytest.raises((FileNotFoundError, TypeError, ValueError)):
        arr.open()

    arr.create()
    assert zarr_path.exists()
    arr.open()


@pytest.mark.skipif(
    ZARR_PYTHON_V3, reason="setting zarr compressor not yet possible for Zarr Python v3"
)
@pytest.mark.parametrize(
    "compressor",
    [
        None,
        {"id": "zstd", "level": 1},
        {"id": "blosc", "cname": "lz4", "clevel": 2, "shuffle": -1},
    ],
)
def test_compression(tmp_path, compressor):
    zarr_path = tmp_path / "lazy.zarr"

    arr = lazy_zarr_array(
        zarr_path, shape=(3, 3), dtype=int, chunks=(2, 2), compressor=compressor
    )
    arr.create()

    # open with zarr python (for zarr python v2 and tensorstore)
    with config.set({"storage_name": "zarr-python"}):
        z = open_backend_array(zarr_path, mode="r")

    if compressor is None:
        assert z.compressor is None
    else:
        assert z.compressor == get_codec(compressor)



import logging

import networkx as nx

from cubed.primitive.blockwise import (
    can_fuse_multiple_primitive_ops,
    can_fuse_primitive_ops,
    fuse,
    fuse_multiple,
)

logger = logging.getLogger(__name__)

DEFAULT_MAX_TOTAL_SOURCE_ARRAYS = 4
DEFAULT_MAX_TOTAL_NUM_INPUT_BLOCKS = 10


def simple_optimize_dag(dag, array_names=None):
    """Apply map blocks fusion."""

    # note there is no need to prune the dag, since the way it is built
    # ensures that only the transitive dependencies of the target arrays are included

    dag = dag.copy()
    nodes = {n: d for (n, d) in dag.nodes(data=True)}

    def can_fuse(n):
        # fuse a single chain looking like this:
        # op1 -> op2_input -> op2

        op2 = n

        # if node (op2) does not have a primitive op then it can't be fused
        if "primitive_op" not in nodes[op2]:
            return False

        # if node (op2) does not have exactly one input and output then don't fuse
        # (it could have no inputs or multiple inputs)
        if dag.in_degree(op2) != 1 or dag.out_degree(op2) != 1:
            return False

        # if input is one of the arrays being computed then don't fuse
        op2_input = next(dag.predecessors(op2))
        if array_names is not None and op2_input in array_names:
            return False

        # if input is used by another node then don't fuse
        if dag.out_degree(op2_input) != 1:
            return False

        # if node producing input (op1) has more than one output then don't fuse
        op1 = next(dag.predecessors(op2_input))
        if dag.out_degree(op1) != 1:
            return False

        # op1 and op2 must have primitive ops that can be fused
        if "primitive_op" not in nodes[op1]:
            return False
        return can_fuse_primitive_ops(
            nodes[op1]["primitive_op"], nodes[op2]["primitive_op"]
        )

    for n in list(dag.nodes()):
        if can_fuse(n):
            op2 = n
            op2_input = next(dag.predecessors(op2))
            op1 = next(dag.predecessors(op2_input))
            op1_inputs = list(dag.predecessors(op1))

            primitive_op = fuse(nodes[op1]["primitive_op"], nodes[op2]["primitive_op"])
            nodes[op2]["primitive_op"] = primitive_op
            nodes[op2]["pipeline"] = primitive_op.pipeline

            for n in op1_inputs:
                dag.add_edge(n, op2)
            dag.remove_node(op2_input)
            dag.remove_node(op1)

    return dag


sym_counter = 0


def gensym(name="op"):
    global sym_counter
    sym_counter += 1
    return f"{name}-{sym_counter:03}"


def predecessors_unordered(dag, name):
    """Return a node's predecessors in no particular order, with repeats for multiple edges."""
    for pre, _ in dag.in_edges(name):
        yield pre


def successors_unordered(dag, name):
    """Return a node's successors in no particular order, with repeats for multiple edges."""
    for _, succ in dag.out_edges(name):
        yield succ


def predecessor_ops(dag, name):
    """Return an op node's op predecessors in the same order as the input source arrays for the op.

    Note that each input source array is produced by a single predecessor op.
    """
    nodes = dict(dag.nodes(data=True))
    for input in nodes[name]["primitive_op"].source_array_names:
        pre_list = list(predecessors_unordered(dag, input))
        assert len(pre_list) == 1  # each array is produced by a single op
        yield pre_list[0]


def predecessor_ops_and_arrays(dag, name):
    # returns op predecessors, the arrays that they produce (only one since we don't support multiple outputs yet),
    # and a flag indicating if the op can be fused with each predecessor, taking into account the number of dependents for the array
    nodes = dict(dag.nodes(data=True))
    for input in nodes[name]["primitive_op"].source_array_names:
        pre_list = list(predecessors_unordered(dag, input))
        assert len(pre_list) == 1  # each array is produced by a single op
        pre = pre_list[0]
        can_fuse = is_primitive_op(nodes[pre]) and out_degree_unique(dag, input) == 1
        yield pre, input, can_fuse


def out_degree_unique(dag, name):
    """Returns number of unique out edges"""
    return len(set(post for _, post in dag.out_edges(name)))


def is_primitive_op(node_dict):
    """Return True if a node is a primitive op"""
    return "primitive_op" in node_dict


def is_fusable_with_predecessors(node_dict):
    """Return True if a node is a primitive op and can be fused with its predecessors."""
    return (
        is_primitive_op(node_dict)
        and node_dict["primitive_op"].fusable_with_predecessors
    )


def num_source_arrays(dag, name):
    """Return the number of (non-hidden) arrays that are inputs to an op.

    Hidden arrays are used for internal bookkeeping, are very small virtual arrays
    (empty, or offsets for example), and are not shown on the plan visualization.
    For these reasons they shouldn't count towards ``max_total_source_arrays``.
    """
    nodes = dict(dag.nodes(data=True))
    return sum(
        not nodes[array]["hidden"] for array in predecessors_unordered(dag, name)
    )


def can_fuse_predecessors(
    dag,
    name,
    *,
    array_names=None,
    max_total_source_arrays=DEFAULT_MAX_TOTAL_SOURCE_ARRAYS,
    max_total_num_input_blocks=DEFAULT_MAX_TOTAL_NUM_INPUT_BLOCKS,
    always_fuse=None,
    never_fuse=None,
):
    nodes = dict(dag.nodes(data=True))

    # if node itself can't be fused then there is nothing to fuse
    if not is_fusable_with_predecessors(nodes[name]):
        logger.debug(
            "can't fuse %s since it is not a primitive operation, or it uses an operation that can't be fused (concat or stack)",
            name,
        )
        return False

    # if no predecessor ops can be fused then there is nothing to fuse
    # (this may be because predecessor ops produce arrays with multiple dependents)
    if all(not can_fuse for _, _, can_fuse in predecessor_ops_and_arrays(dag, name)):
        logger.debug("can't fuse %s since no predecessor ops can be fused", name)
        return False

    # if a predecessor op produces one of the arrays being computed, then don't fuse
    if array_names is not None:
        predecessor_array_names = set(
            array_name for _, array_name, _ in predecessor_ops_and_arrays(dag, name)
        )
        array_names_intersect = set(array_names) & predecessor_array_names
        if len(array_names_intersect) > 0:
            logger.debug(
                "can't fuse %s since predecessor ops produce one or more arrays being computed %s",
                name,
                array_names_intersect,
            )
            return False

    # if any predecessor ops have multiple outputs then don't fuse
    # TODO: implement "child fusion" (where a multiple output op fuses its children)
    if any(
        len(list(successors_unordered(dag, pre))) > 1
        for pre in predecessor_ops(dag, name)
    ):
        logger.debug(
            "can't fuse %s since at least one predecessor has multiple outputs", name
        )
        return False

    # if node is in never_fuse or always_fuse list then it overrides logic below
    if never_fuse is not None and name in never_fuse:
        logger.debug("can't fuse %s since it is in 'never_fuse'", name)
        return False
    if always_fuse is not None and name in always_fuse:
        logger.debug("can fuse %s since it is in 'always_fuse'", name)
        return True

    # if there is more than a single predecessor op, and the total number of source arrays to
    # the fused function would be more than an allowed maximum, then don't fuse
    if len(list(predecessor_ops(dag, name))) > 1:
        total_source_arrays = sum(
            num_source_arrays(dag, pre) if can_fuse else 1
            for pre, _, can_fuse in predecessor_ops_and_arrays(dag, name)
        )
        if total_source_arrays > max_total_source_arrays:
            logger.debug(
                "can't fuse %s since total number of source arrays (%s) exceeds max (%s)",
                name,
                total_source_arrays,
                max_total_source_arrays,
            )
            return False

    # if a predecessor has no primitive op then just use None
    predecessor_primitive_ops = [
        nodes[pre]["primitive_op"] if can_fuse else None
        for pre, _, can_fuse in predecessor_ops_and_arrays(dag, name)
    ]
    return can_fuse_multiple_primitive_ops(
        name,
        nodes[name]["primitive_op"],
        predecessor_primitive_ops,
        max_total_num_input_blocks=max_total_num_input_blocks,
    )


def fuse_predecessors(
    dag,
    name,
    *,
    array_names=None,
    max_total_source_arrays=DEFAULT_MAX_TOTAL_SOURCE_ARRAYS,
    max_total_num_input_blocks=DEFAULT_MAX_TOTAL_NUM_INPUT_BLOCKS,
    always_fuse=None,
    never_fuse=None,
):
    """Fuse a node with its immediate predecessors."""

    # if can't fuse then return dag unchanged
    if not can_fuse_predecessors(
        dag,
        name,
        array_names=array_names,
        max_total_source_arrays=max_total_source_arrays,
        max_total_num_input_blocks=max_total_num_input_blocks,
        always_fuse=always_fuse,
        never_fuse=never_fuse,
    ):
        return dag

    nodes = dict(dag.nodes(data=True))

    primitive_op = nodes[name]["primitive_op"]

    # if a predecessor has no primitive op then just use None
    predecessor_primitive_ops = [
        nodes[pre]["primitive_op"] if can_fuse else None
        for pre, _, can_fuse in predecessor_ops_and_arrays(dag, name)
    ]

    fused_primitive_op = fuse_multiple(primitive_op, *predecessor_primitive_ops)

    fused_dag = dag.copy()
    fused_nodes = dict(fused_dag.nodes(data=True))

    fused_nodes[name]["primitive_op"] = fused_primitive_op
    fused_nodes[name]["pipeline"] = fused_primitive_op.pipeline

    # re-wire dag to remove predecessor nodes that have been fused
    for pre, input, can_fuse in predecessor_ops_and_arrays(dag, name):
        if can_fuse:
            # check if already removed for repeated arguments
            if input in fused_dag:
                fused_dag.remove_node(input)
            if pre in fused_dag:
                fused_dag.remove_node(pre)
            for pre_input in predecessors_unordered(dag, pre):
                fused_dag.add_edge(pre_input, name)

    return fused_dag


def multiple_inputs_optimize_dag(
    dag,
    *,
    array_names=None,
    max_total_source_arrays=DEFAULT_MAX_TOTAL_SOURCE_ARRAYS,
    max_total_num_input_blocks=DEFAULT_MAX_TOTAL_NUM_INPUT_BLOCKS,
    always_fuse=None,
    never_fuse=None,
):
    """Fuse multiple inputs."""
    for name in list(nx.topological_sort(dag)):
        if name.startswith("array-"):
            continue
        dag = fuse_predecessors(
            dag,
            name,
            array_names=array_names,
            max_total_source_arrays=max_total_source_arrays,
            max_total_num_input_blocks=max_total_num_input_blocks,
            always_fuse=always_fuse,
            never_fuse=never_fuse,
        )
    return dag


def fuse_all_optimize_dag(dag, array_names=None):
    """Force all operations to be fused."""
    dag = dag.copy()
    always_fuse = [op for op in dag.nodes() if op.startswith("op-")]
    return multiple_inputs_optimize_dag(
        dag, array_names=array_names, always_fuse=always_fuse
    )


def fuse_only_optimize_dag(dag, *, array_names=None, only_fuse=None):
    """Force only specified operations to be fused, all others will be left even if they are suitable for fusion."""
    dag = dag.copy()
    always_fuse = only_fuse
    never_fuse = set(op for op in dag.nodes() if op.startswith("op-")) - set(only_fuse)
    return multiple_inputs_optimize_dag(
        dag, array_names=array_names, always_fuse=always_fuse, never_fuse=never_fuse
    )



import numpy as np
import pytest

import cubed
import cubed.array_api as xp
from cubed._testing import assert_array_equal


@pytest.fixture
def spec(tmp_path):
    return cubed.Spec(tmp_path, allowed_mem=100000)


def test_pad(spec):
    an = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])

    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2), spec=spec)
    b = cubed.pad(a, ((1, 0), (0, 0)), mode="symmetric")
    assert b.chunks == ((2, 2), (2, 1))

    assert_array_equal(b.compute(), np.pad(an, ((1, 0), (0, 0)), mode="symmetric"))



import os
import platform
import re

import numpy as np
import psutil
import pytest

import cubed
import cubed.array_api as xp
import cubed.random
from cubed._testing import assert_array_equal
from cubed.diagnostics import ProgressBar
from cubed.diagnostics.history import HistoryCallback
from cubed.diagnostics.mem_usage import MemoryVisualizationCallback
from cubed.diagnostics.mem_warn import MemoryWarningCallback
from cubed.diagnostics.rich import RichProgressBar
from cubed.diagnostics.timeline import TimelineVisualizationCallback
from cubed.diagnostics.tqdm import TqdmProgressBar
from cubed.primitive.blockwise import apply_blockwise
from cubed.runtime.create import create_executor
from cubed.storage.backend import backend_storage_name
from cubed.tests.utils import (
    ALL_EXECUTORS,
    MAIN_EXECUTORS,
    MODAL_EXECUTORS,
    TaskCounter,
)


@pytest.fixture
def spec(tmp_path):
    return cubed.Spec(tmp_path, allowed_mem=100000)


@pytest.fixture(
    scope="module",
    params=MAIN_EXECUTORS,
    ids=[executor.name for executor in MAIN_EXECUTORS],
)
def executor(request):
    return request.param


@pytest.fixture(
    scope="module",
    params=ALL_EXECUTORS,
    ids=[executor.name for executor in ALL_EXECUTORS],
)
def any_executor(request):
    return request.param


@pytest.fixture(
    scope="module",
    params=MODAL_EXECUTORS,
    ids=[executor.name for executor in MODAL_EXECUTORS],
)
def modal_executor(request):
    return request.param


mock_call_counter = 0


def mock_apply_blockwise(*args, **kwargs):
    # Raise an error on every 3rd call
    global mock_call_counter
    mock_call_counter += 1
    if mock_call_counter % 3 == 0:
        raise IOError("Test fault injection")
    return apply_blockwise(*args, **kwargs)


# see tests/runtime for more tests for retries for other executors
@pytest.mark.skipif(
    platform.system() == "Windows", reason="measuring memory does not run on windows"
)
def test_retries(mocker, spec):
    # Use threads executor since single-threaded executor doesn't support retries
    executor = create_executor("threads")
    # Inject faults into the primitive layer
    mocker.patch(
        "cubed.primitive.blockwise.apply_blockwise", side_effect=mock_apply_blockwise
    )

    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2), spec=spec)
    b = xp.asarray([[1, 1, 1], [1, 1, 1], [1, 1, 1]], chunks=(2, 2), spec=spec)
    c = xp.add(a, b)
    assert_array_equal(
        c.compute(executor=executor), np.array([[2, 3, 4], [5, 6, 7], [8, 9, 10]])
    )


@pytest.mark.skipif(
    platform.system() == "Windows", reason="measuring memory does not run on windows"
)
def test_callbacks(spec, executor):
    task_counter = TaskCounter()
    # test following indirectly by checking they don't cause a failure
    progress = TqdmProgressBar()
    hist = HistoryCallback()
    timeline_viz = TimelineVisualizationCallback()
    memory_viz = MemoryVisualizationCallback()

    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2), spec=spec)
    b = xp.asarray([[1, 1, 1], [1, 1, 1], [1, 1, 1]], chunks=(2, 2), spec=spec)
    c = xp.add(a, b)
    assert_array_equal(
        c.compute(
            executor=executor,
            callbacks=[task_counter, progress, hist, timeline_viz, memory_viz],
        ),
        np.array([[2, 3, 4], [5, 6, 7], [8, 9, 10]]),
    )

    num_created_arrays = 1
    assert task_counter.value == num_created_arrays + 4


def test_callbacks_as_context_managers(spec, executor):
    with TaskCounter() as task_counter, ProgressBar():
        assert task_counter is not None
        a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2), spec=spec)
        b = xp.asarray([[1, 1, 1], [1, 1, 1], [1, 1, 1]], chunks=(2, 2), spec=spec)
        c = xp.add(a, b)
        assert_array_equal(
            c.compute(executor=executor),
            np.array([[2, 3, 4], [5, 6, 7], [8, 9, 10]]),
        )

        num_created_arrays = 1
        assert task_counter.value == num_created_arrays + 4


def test_rich_progress_bar(spec, executor):
    # test indirectly by checking it doesn't cause a failure
    progress = RichProgressBar()

    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2), spec=spec)
    b = xp.asarray([[1, 1, 1], [1, 1, 1], [1, 1, 1]], chunks=(2, 2), spec=spec)
    c = xp.add(a, b)
    assert_array_equal(
        c.compute(executor=executor, callbacks=[progress]),
        np.array([[2, 3, 4], [5, 6, 7], [8, 9, 10]]),
    )


@pytest.mark.cloud
def test_callbacks_modal(spec, modal_executor):
    task_counter = TaskCounter(check_timestamps=False)
    tmp_path = "s3://cubed-unittest/callbacks"
    spec = cubed.Spec(
        tmp_path, allowed_mem=100000, storage_options=dict(use_obstore=True)
    )

    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2), spec=spec)
    b = xp.asarray([[1, 1, 1], [1, 1, 1], [1, 1, 1]], chunks=(2, 2), spec=spec)
    c = xp.add(a, b)
    assert_array_equal(
        c.compute(executor=modal_executor, callbacks=[task_counter]),
        np.array([[2, 3, 4], [5, 6, 7], [8, 9, 10]]),
    )

    num_created_arrays = 1
    assert task_counter.value == num_created_arrays + 4


@pytest.mark.skipif(
    platform.system() == "Windows", reason="measuring memory does not run on windows"
)
def test_mem_warn(tmp_path, executor):
    if executor.name not in ("processes", "lithops"):
        pytest.skip(f"{executor.name} executor does not support MemoryWarningCallback")

    spec = cubed.Spec(tmp_path, allowed_mem=200_000_000, reserved_mem=100_000_000)
    mem_warn = MemoryWarningCallback()

    def func(a):
        np.ones(100_000_000)  # blow memory
        return a

    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2), spec=spec)
    b = cubed.map_blocks(func, a, dtype=a.dtype)
    with pytest.raises(
        UserWarning, match="Peak memory usage exceeded allowed_mem when running tasks"
    ):
        b.compute(executor=executor, callbacks=[mem_warn])


@pytest.mark.skipif(
    backend_storage_name() == "tensorstore",
    reason="tensorstore does not support resume",
)
def test_resume(spec, executor):
    if executor.name == "beam":
        pytest.skip(f"{executor.name} executor does not support resume")

    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2), spec=spec)
    b = xp.asarray([[1, 1, 1], [1, 1, 1], [1, 1, 1]], chunks=(2, 2), spec=spec)
    c = xp.add(a, b)
    d = xp.negative(c)

    num_created_arrays = 2  # c, d
    assert d.plan._finalize(optimize_graph=False).num_tasks() == num_created_arrays + 8

    task_counter = TaskCounter()
    c.compute(executor=executor, callbacks=[task_counter], optimize_graph=False)
    num_created_arrays = 1  # c
    assert task_counter.value == num_created_arrays + 4

    # since c has already been computed, when computing d only 4 tasks are run, instead of 8
    task_counter = TaskCounter()
    d.compute(
        executor=executor,
        callbacks=[task_counter],
        optimize_graph=False,
        resume=True,
    )
    # the create arrays tasks are run again, even though they exist
    num_created_arrays = 2  # c, d
    assert task_counter.value == num_created_arrays + 4


@pytest.mark.parametrize("compute_arrays_in_parallel", [True, False])
def test_compute_arrays_in_parallel(spec, any_executor, compute_arrays_in_parallel):
    if any_executor.name == "beam":
        pytest.skip(
            f"{any_executor.name} executor does not support compute_arrays_in_parallel"
        )

    a = cubed.random.random((10, 10), chunks=(5, 5), spec=spec)
    b = cubed.random.random((10, 10), chunks=(5, 5), spec=spec)
    c = xp.add(a, b)

    # note that this merely checks that compute_arrays_in_parallel is accepted
    c.compute(
        executor=any_executor, compute_arrays_in_parallel=compute_arrays_in_parallel
    )


@pytest.mark.cloud
@pytest.mark.parametrize("compute_arrays_in_parallel", [True, False])
def test_compute_arrays_in_parallel_modal(modal_executor, compute_arrays_in_parallel):
    tmp_path = "s3://cubed-unittest/parallel_pipelines"
    spec = cubed.Spec(
        tmp_path, allowed_mem=100000, storage_options=dict(use_obstore=True)
    )

    a = cubed.random.random((10, 10), chunks=(5, 5), spec=spec)
    b = cubed.random.random((10, 10), chunks=(5, 5), spec=spec)
    c = xp.add(a, b)

    # note that this merely checks that compute_arrays_in_parallel is accepted
    c.compute(
        executor=modal_executor,
        compute_arrays_in_parallel=compute_arrays_in_parallel,
    )


def test_check_runtime_memory_dask(spec, executor):
    if executor.name != "dask":
        pytest.skip(f"{executor.name} executor does not support check_runtime_memory")

    spec = cubed.Spec(spec.work_dir, allowed_mem="4GB")  # larger than runtime memory
    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2), spec=spec)
    b = xp.asarray([[1, 1, 1], [1, 1, 1], [1, 1, 1]], chunks=(2, 2), spec=spec)
    c = xp.add(a, b)
    with pytest.raises(
        ValueError,
        match=r"Runtime memory \(2000000000\) is less than allowed_mem \(4000000000\)",
    ):
        c.compute(
            executor=executor,
            compute_kwargs=dict(memory_limit="2 GB", threads_per_worker=1),
        )


def test_check_runtime_memory_dask_no_workers(spec, executor):
    if executor.name != "dask":
        pytest.skip(f"{executor.name} executor does not support check_runtime_memory")

    spec = cubed.Spec(spec.work_dir, allowed_mem=100000)
    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2), spec=spec)
    b = xp.asarray([[1, 1, 1], [1, 1, 1], [1, 1, 1]], chunks=(2, 2), spec=spec)
    c = xp.add(a, b)
    with pytest.raises(ValueError, match=r"Cluster has no workers running"):
        c.compute(
            executor=executor,
            compute_kwargs=dict(n_workers=0),
        )


@pytest.mark.cloud
def test_check_runtime_memory_modal(spec, modal_executor):
    tmp_path = "s3://cubed-unittest/check-runtime-memory"
    spec = cubed.Spec(
        tmp_path,
        allowed_mem="4GB",  # larger than Modal runtime memory
        storage_options=dict(use_obstore=True),
    )
    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2), spec=spec)
    b = xp.asarray([[1, 1, 1], [1, 1, 1], [1, 1, 1]], chunks=(2, 2), spec=spec)
    c = xp.add(a, b)
    with pytest.raises(
        ValueError,
        match=r"Runtime memory \(2000000000\) is less than allowed_mem \(4000000000\)",
    ):
        c.compute(executor=modal_executor)


def test_check_runtime_memory_processes(spec, executor):
    if executor.name != "processes":
        pytest.skip(f"{executor.name} executor does not support check_runtime_memory")

    total_mem = psutil.virtual_memory().total
    max_workers = os.cpu_count()
    mem_per_worker = total_mem // max_workers
    allowed_mem = mem_per_worker * 2  # larger than will fit

    spec = cubed.Spec(spec.work_dir, allowed_mem=allowed_mem)
    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2), spec=spec)
    b = xp.asarray([[1, 1, 1], [1, 1, 1], [1, 1, 1]], chunks=(2, 2), spec=spec)
    c = xp.add(a, b)
    with pytest.raises(
        ValueError,
        match=re.escape(
            f"Total memory on machine ({total_mem}) is less than allowed_mem * max_workers ({allowed_mem} * {max_workers} = {allowed_mem * max_workers})"
        ),
    ):
        c.compute(executor=executor)

    # OK if we use fewer workers
    c.compute(executor=executor, max_workers=max_workers // 2)


COMPILE_FUNCTIONS = [lambda fn: fn]

try:
    from numba import jit as numba_jit

    COMPILE_FUNCTIONS.append(numba_jit)
except ModuleNotFoundError:
    pass

try:
    if "jax" in os.environ.get("CUBED_BACKEND_ARRAY_API_MODULE", ""):
        from jax import jit as jax_jit

        COMPILE_FUNCTIONS.append(jax_jit)
except ModuleNotFoundError:
    pass


@pytest.mark.parametrize("compile_function", COMPILE_FUNCTIONS)
def test_check_compilation(spec, executor, compile_function):
    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2), spec=spec)
    b = xp.asarray([[1, 1, 1], [1, 1, 1], [1, 1, 1]], chunks=(2, 2), spec=spec)
    c = xp.add(a, b)
    assert_array_equal(
        c.compute(executor=executor, compile_function=compile_function),
        np.array([[2, 3, 4], [5, 6, 7], [8, 9, 10]]),
    )


def test_compilation_can_fail(spec, executor):
    def compile_function(func):
        raise NotImplementedError(f"Cannot compile {func}")

    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2), spec=spec)
    b = xp.asarray([[1, 1, 1], [1, 1, 1], [1, 1, 1]], chunks=(2, 2), spec=spec)
    c = xp.add(a, b)
    with pytest.raises(NotImplementedError) as excinfo:
        c.compute(executor=executor, compile_function=compile_function)

    assert "add" in str(excinfo.value), "Compile function was applied to add operation."


def test_compilation_with_config_can_fail(spec, executor):
    def compile_function(func, *, config=None):
        raise NotImplementedError(f"Cannot compile {func} with {config}")

    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2), spec=spec)
    b = xp.asarray([[1, 1, 1], [1, 1, 1], [1, 1, 1]], chunks=(2, 2), spec=spec)
    c = xp.add(a, b)
    with pytest.raises(NotImplementedError) as excinfo:
        c.compute(executor=executor, compile_function=compile_function)

    assert "BlockwiseSpec" in str(excinfo.value), (
        "Compile function was applied with a config argument."
    )



import asyncio
from typing import Any, Callable, Dict, Optional, Sequence

from dask.distributed import Client
from networkx import MultiDiGraph

from cubed.runtime.asyncio import async_map_dag
from cubed.runtime.backup import use_backups_default
from cubed.runtime.types import Callback, DagExecutor
from cubed.runtime.utils import asyncio_run, execution_stats, gensym
from cubed.spec import Spec


# note we can't call `pipeline_func` just `func` here as it clashes with `dask.distributed.Client.map``
@execution_stats
def run_func(input, pipeline_func=None, config=None, name=None, compute_id=None):
    result = pipeline_func(input, config=config)
    return result


def check_runtime_memory(spec, client):
    allowed_mem = spec.allowed_mem if spec is not None else None
    scheduler_info = client.scheduler_info()
    workers = scheduler_info.get("workers")
    if workers is None or len(workers) == 0:
        raise ValueError("Cluster has no workers running")
    runtime_memory = min(w["memory_limit"] // w["nthreads"] for w in workers.values())
    if allowed_mem is not None:
        if runtime_memory < allowed_mem:
            raise ValueError(
                f"Runtime memory ({runtime_memory}) is less than allowed_mem ({allowed_mem})"
            )


def dask_create_futures_func(
    client,
    function: Callable[..., Any],
    name: Optional[str] = None,
    retries: Optional[str] = None,
):
    def create_futures_func(input, **kwargs):
        input = list(input)  # dask expects a sequence (it calls `len` on it)
        key = name or gensym("map")
        key = key.replace("-", "_")  # otherwise array number is not shown on dashboard
        if "func" in kwargs:
            kwargs["pipeline_func"] = kwargs.pop("func")  # rename to avoid clash
        return [
            (i, asyncio.ensure_future(f))
            for i, f in zip(
                input,
                client.map(function, input, key=key, retries=retries, **kwargs),
            )
        ]

    return create_futures_func


class DaskExecutor(DagExecutor):
    """An execution engine that uses Dask Distributed's async API."""

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    @property
    def name(self) -> str:
        return "dask"

    def execute_dag(
        self,
        dag: MultiDiGraph,
        callbacks: Optional[Sequence[Callback]] = None,
        spec: Optional[Spec] = None,
        compute_id: Optional[str] = None,
        compute_kwargs: Optional[Dict[str, Any]] = None,
        **kwargs,
    ) -> None:
        merged_kwargs = {**self.kwargs, **kwargs}
        asyncio_run(
            self._async_execute_dag(
                dag,
                callbacks=callbacks,
                spec=spec,
                compute_kwargs=compute_kwargs,
                compute_id=compute_id,
                **merged_kwargs,
            )
        )

    async def _async_execute_dag(
        self,
        dag: MultiDiGraph,
        callbacks: Optional[Sequence[Callback]] = None,
        spec: Optional[Spec] = None,
        compute_kwargs: Optional[Dict[str, Any]] = None,
        compute_arrays_in_parallel: Optional[bool] = None,
        **kwargs,
    ) -> None:
        compute_kwargs = compute_kwargs or {}
        retries = kwargs.pop("retries", 2)
        name = kwargs.get("name", None)
        async with Client(asynchronous=True, **compute_kwargs) as client:
            if spec is not None:
                check_runtime_memory(spec, client)
                if "use_backups" not in kwargs and use_backups_default(spec):
                    kwargs["use_backups"] = True

            create_futures_func = dask_create_futures_func(
                client, run_func, name, retries=retries
            )
            create_backup_futures_func = dask_create_futures_func(
                client, run_func, name
            )

            await async_map_dag(
                create_futures_func,
                dag=dag,
                callbacks=callbacks,
                compute_arrays_in_parallel=compute_arrays_in_parallel,
                create_backup_futures_func=create_backup_futures_func,
                **kwargs,
            )



import random as pyrandom

from numpy.random import Generator, Philox

from cubed.backend_array_api import namespace as nxp
from cubed.backend_array_api import numpy_array_to_backend_array
from cubed.core.ops import map_blocks
from cubed.utils import block_id_to_offset, normalize_chunks, normalize_shape


def random(size, *, dtype=nxp.float64, chunks=None, spec=None):
    """Return random floats in the half-open interval [0.0, 1.0)."""
    shape = normalize_shape(size)
    chunks = normalize_chunks(chunks, shape=shape, dtype=dtype)
    numblocks = tuple(map(len, chunks))
    root_seed = pyrandom.getrandbits(128)
    extra_func_kwargs = dict(dtype=dtype)
    return map_blocks(
        _random,
        dtype=dtype,
        chunks=chunks,
        spec=spec,
        numblocks=numblocks,
        root_seed=root_seed,
        extra_func_kwargs=extra_func_kwargs,
    )


def _random(x, numblocks=None, root_seed=None, dtype=nxp.float64, block_id=None):
    stream_id = block_id_to_offset(block_id, numblocks)
    rg = Generator(Philox(key=root_seed + stream_id))
    out = rg.random(x.shape, dtype=dtype)
    out = numpy_array_to_backend_array(out)
    return out



from pathlib import Path
from typing import Literal, Optional

import zarr

from cubed.types import T_DType, T_RegularChunks, T_Shape, T_Store

# always write empty chunks to avoid a check for each chunk
zarr.config.set({"array.write_empty_chunks": True})


class ZarrV3ArrayGroup(dict):
    def __init__(
        self,
        shape: Optional[T_Shape] = None,
        dtype: Optional[T_DType] = None,
        chunks: Optional[T_RegularChunks] = None,
    ):
        dict.__init__(self)
        self.shape = shape
        self.dtype = dtype
        self.chunks = chunks

    def __getitem__(self, key):
        if isinstance(key, str):
            return super().__getitem__(key)
        return {field: zarray[key] for field, zarray in self.items()}

    def set_basic_selection(self, selection, value, fields=None):
        self[fields][selection] = value


def open_zarr_v3_array(
    store: T_Store,
    mode: Optional[Literal["r", "r+", "a", "w", "w-"]],
    *,
    shape: Optional[T_Shape] = None,
    dtype: Optional[T_DType] = None,
    chunks: Optional[T_RegularChunks] = None,
    path: Optional[str] = None,
    **kwargs,
):
    # use obstore if requested
    storage_options = kwargs.get("storage_options", None)
    if storage_options is not None and storage_options.get("use_obstore", False):
        import obstore as obs
        from zarr.storage import ObjectStore

        if isinstance(store, str):
            if "://" not in store:
                p = Path(store)
                store = ObjectStore(obs.store.from_url(p.as_uri(), mkdir=True))
            else:
                store = ObjectStore(obs.store.from_url(store))
        elif isinstance(store, Path):
            p = store
            store = ObjectStore(obs.store.from_url(p.as_uri(), mkdir=True))

    if isinstance(chunks, int):
        chunks = (chunks,)

    if dtype is None or not hasattr(dtype, "fields") or dtype.fields is None:
        return zarr.open(
            store=store,
            mode=mode,
            shape=shape,
            dtype=dtype,
            chunks=chunks,
            path=path,
        )

    assert mode is not None
    group = zarr.open_group(store=store, mode=mode, path=path)

    # create/open all the arrays in the group
    ret = ZarrV3ArrayGroup(shape=shape, dtype=dtype, chunks=chunks)
    for field in dtype.fields:
        field_dtype, _ = dtype.fields[field]
        if mode in ("r", "r+"):
            ret[field] = group[field]
        else:
            assert chunks is not None
            ret[field] = group.create_array(
                field,
                shape=shape,
                dtype=field_dtype,
                chunks=chunks,
            )
    return ret



# flake8: noqa
from .array import CoreArray, compute, gensym, measure_reserved_mem, visualize
from .gufunc import apply_gufunc
from .ops import (
    blockwise,
    elemwise,
    from_array,
    from_zarr,
    map_blocks,
    rechunk,
    reduction,
    scan,
    squeeze,
    store,
    to_zarr,
    unify_chunks,
)
from .plan import Plan



from cubed.backend_array_api import namespace as nxp


class __array_namespace_info__:
    def capabilities(self):
        return {
            "boolean indexing": False,  # not supported in Cubed (#73)
            "data-dependent shapes": False,  # not supported in Cubed
            "max dimensions": nxp.__array_namespace_info__().capabilities()[
                "max dimensions"
            ],
        }

    # devices and dtypes are determined by the backend array API

    def default_device(self):
        return nxp.__array_namespace_info__().default_device()

    def default_dtypes(self, *, device=None):
        return nxp.__array_namespace_info__().default_dtypes(device=device)

    def devices(self):
        return nxp.__array_namespace_info__().devices()

    def dtypes(self, *, device=None, kind=None):
        return nxp.__array_namespace_info__().dtypes(device=device, kind=kind)



import dataclasses
import math
from typing import Any, Dict, Optional, Union

import numpy as np
import tensorstore

from cubed.types import T_DType, T_RegularChunks, T_Shape, T_Store
from cubed.utils import join_path


@dataclasses.dataclass(frozen=True)
class TensorStoreArray:
    array: tensorstore.TensorStore

    @property
    def shape(self) -> tuple[int, ...]:
        return self.array.shape

    @property
    def dtype(self) -> np.dtype:
        return self.array.dtype.numpy_dtype

    @property
    def chunks(self) -> tuple[int, ...]:
        return self.array.chunk_layout.read_chunk.shape or ()

    @property
    def ndim(self) -> int:
        return len(self.shape)

    @property
    def size(self) -> int:
        return math.prod(self.shape)

    @property
    def oindex(self):
        return self.array.oindex

    def __getitem__(self, key):
        # read eagerly
        return self.array.__getitem__(key).read().result()

    def __setitem__(self, key, value):
        self.array.__setitem__(key, value)


class TensorStoreGroup(dict):
    def __init__(
        self,
        shape: Optional[T_Shape] = None,
        dtype: Optional[T_DType] = None,
        chunks: Optional[T_RegularChunks] = None,
    ):
        dict.__init__(self)
        self.shape = shape
        self.dtype = dtype
        self.chunks = chunks

    def __getitem__(self, key):
        if isinstance(key, str):
            return super().__getitem__(key)
        return {field: zarray[key] for field, zarray in self.items()}

    def set_basic_selection(self, selection, value, fields=None):
        self[fields][selection] = value


def encode_dtype(d):
    if d.fields is None:
        return d.str
    else:
        return d.descr


def get_metadata(dtype, chunks, compressor):
    metadata = {}
    if dtype is not None:
        dtype = np.dtype(dtype)
        metadata["dtype"] = encode_dtype(dtype)
    if chunks is not None:
        if isinstance(chunks, int):
            chunks = (chunks,)
        metadata["chunks"] = chunks
    if compressor != "default":
        metadata["compressor"] = compressor
    return metadata


def open_tensorstore_array(
    store: T_Store,
    mode: str,
    *,
    shape: Optional[T_Shape] = None,
    dtype: Optional[T_DType] = None,
    chunks: Optional[T_RegularChunks] = None,
    path: Optional[str] = None,
    compressor: Union[dict, str, None] = "default",
    **kwargs,
):
    store = str(store)  # TODO: check if Path or str

    spec: Dict[str, Any]
    if "://" in store:
        spec = {"driver": "zarr", "kvstore": store}
    else:
        spec = {
            "driver": "zarr",
            "kvstore": {"driver": "file", "path": store},
            "path": path or "",
        }

    if mode == "r":
        open_kwargs = dict(read=True, open=True)
    elif mode == "r+":
        open_kwargs = dict(read=True, write=True, open=True)
    elif mode == "a":
        open_kwargs = dict(read=True, write=True, open=True, create=True)
    elif mode == "w":
        open_kwargs = dict(read=True, write=True, create=True, delete_existing=True)
    elif mode == "w-":
        open_kwargs = dict(read=True, write=True, create=True)
    else:
        raise ValueError(f"Mode not supported: {mode}")

    if dtype is None or not hasattr(dtype, "fields") or dtype.fields is None:
        metadata = get_metadata(dtype, chunks, compressor)
        if metadata:
            spec["metadata"] = metadata

        return TensorStoreArray(
            tensorstore.open(
                spec,
                shape=shape,
                dtype=dtype,
                **open_kwargs,
            ).result()
        )
    else:
        ret = TensorStoreGroup(shape=shape, dtype=dtype, chunks=chunks)
        for field in dtype.fields:
            field_path = field if path is None else join_path(path, field)
            spec["path"] = field_path

            field_dtype, _ = dtype.fields[field]
            metadata = get_metadata(field_dtype, chunks, compressor)
            if metadata:
                spec["metadata"] = metadata

            ret[field] = TensorStoreArray(
                tensorstore.open(
                    spec,
                    shape=shape,
                    dtype=field_dtype,
                    **open_kwargs,
                ).result()
            )
        return ret



import random

import pytest

import cubed
import cubed.array_api as xp
import cubed.random
from cubed._testing import assert_array_equal
from cubed.backend_array_api import namespace as nxp
from cubed.tests.utils import MAIN_EXECUTORS


@pytest.fixture
def spec(tmp_path):
    return cubed.Spec(tmp_path, allowed_mem=100000)


@pytest.fixture(
    scope="module",
    params=MAIN_EXECUTORS,
    ids=[executor.name for executor in MAIN_EXECUTORS],
)
def executor(request):
    return request.param


def test_random(spec, executor):
    a = cubed.random.random((10, 10), chunks=(4, 5), spec=spec)

    assert a.shape == (10, 10)
    assert a.chunks == ((4, 4, 2), (5, 5))
    assert a.dtype == xp.float64

    x = nxp.unique_values(a.compute(executor=executor))
    assert x.dtype == xp.float64
    assert len(x) > 90


def test_random_dtype(spec, executor):
    a = cubed.random.random((10, 10), dtype=xp.float32, chunks=(4, 5), spec=spec)

    assert a.shape == (10, 10)
    assert a.chunks == ((4, 4, 2), (5, 5))
    assert a.dtype == xp.float32

    x = nxp.unique_values(a.compute(executor=executor))
    assert x.dtype == xp.float32
    assert len(x) > 90


def test_random_add(spec, executor):
    a = cubed.random.random((10, 10), chunks=(5, 5), spec=spec)
    b = cubed.random.random((10, 10), chunks=(5, 5), spec=spec)

    c = xp.add(a, b)

    x = nxp.unique_values(c.compute(executor=executor))
    assert len(x) > 90


def test_random_seed(spec, executor):
    random.seed(42)
    a = cubed.random.random((10, 10), chunks=(5, 5), spec=spec)
    a_result = a.compute(executor=executor)

    random.seed(42)
    b = cubed.random.random((10, 10), chunks=(5, 5), spec=spec)
    b_result = b.compute(executor=executor)

    assert_array_equal(a_result, b_result)



import itertools
import uuid
from datetime import datetime

import pytest

modal = pytest.importorskip("modal")

# import MODAL_EXECUTORS to set up config correctly
from cubed.tests.utils import MODAL_EXECUTORS  # isort: skip # noqa: F401

import asyncio

from cubed.runtime.asyncio import async_map_unordered
from cubed.runtime.executors.modal import modal_create_futures_func
from cubed.tests.runtime.utils import check_invocation_counts, deterministic_failure

UNIQUE_DIR_NAME = f"test-{datetime.now().strftime('%Y%m%dT%H%M%S')}-{uuid.uuid4()}"
BASE_PATH = f"s3://cubed-unittest/map_unordered/{UNIQUE_DIR_NAME}"
region = "us-east-1"  # S3 region for above bucket

app = modal.App("cubed-test-app", include_source=True)

image = modal.Image.debian_slim().pip_install(
    [
        "array-api-compat",
        "donfig",
        "mypy_extensions",  # for rechunker
        "ndindex",
        "networkx",
        "obstore",
        "psutil",
        "pytest-mock",  # TODO: only needed for tests
        "tenacity",
        "toolz",
        "zarr",
    ]
)


@app.function(
    image=image,
    secrets=[modal.Secret.from_name("aws-secret-us-east-1")],
    retries=2,
    timeout=10,
    cloud="aws",
    region=region,
)
def deterministic_failure_modal(i, path=None, timing_map=None, *, name=None):
    return deterministic_failure(path, timing_map, i, name=name)


@app.function(
    image=image,
    secrets=[modal.Secret.from_name("aws-secret-us-east-1")],
    timeout=10,
    cloud="aws",
    region=region,
)
def deterministic_failure_modal_no_retries(i, path=None, timing_map=None, *, name=None):
    return deterministic_failure(path, timing_map, i, name=name)


@app.function(
    image=image,
    secrets=[modal.Secret.from_name("aws-secret-us-east-1")],
    retries=2,
    timeout=300,
    cloud="aws",
    region=region,
)
def deterministic_failure_modal_long_timeout(
    i, path=None, timing_map=None, *, name=None
):
    return deterministic_failure(path, timing_map, i, name=name)


async def run_test(app_function, input, use_backups=False, batch_size=None, **kwargs):
    outputs = set()
    with modal.enable_output():
        async with app.run():
            create_futures_func = modal_create_futures_func(app_function)
            async for output in async_map_unordered(
                create_futures_func,
                input,
                use_backups=use_backups,
                batch_size=batch_size,
                **kwargs,
            ):
                outputs.add(output)
    return outputs


# fmt: off
@pytest.mark.parametrize(
    ("timing_map", "n_tasks", "retries"),
    [
        # no failures
        ({}, 3, 2),
        # first invocation fails
        ({0: [-1], 1: [-1], 2: [-1]}, 3, 2),
        # first two invocations fail
        ({0: [-1, -1], 1: [-1, -1], 2: [-1, -1]}, 3, 2),
        # first input sleeps once
        ({0: [20]}, 3, 2),
    ],
)
# fmt: on
@pytest.mark.parametrize("use_backups", [False, True])
@pytest.mark.cloud
def test_success(tmp_path, timing_map, n_tasks, retries, use_backups):
    path = f"{BASE_PATH}/{tmp_path.name}"
    outputs = asyncio.run(
        run_test(
            app_function=deterministic_failure_modal,
            input=range(n_tasks),
            use_backups=use_backups,
            path=path,
            timing_map=timing_map,
        )
    )

    assert outputs == set(range(n_tasks))
    check_invocation_counts(path, timing_map, n_tasks, retries)


# fmt: off
@pytest.mark.parametrize(
    ("timing_map", "n_tasks", "retries"),
    [
        # too many failures
        ({0: [-1], 1: [-1], 2: [-1, -1, -1]}, 3, 2),
    ],
)
# fmt: on
@pytest.mark.parametrize("use_backups", [False, True])
@pytest.mark.cloud
def test_failure(tmp_path, timing_map, n_tasks, retries, use_backups):
    path = f"{BASE_PATH}/{tmp_path.name}"
    with pytest.raises(RuntimeError):
        asyncio.run(
            run_test(
                app_function=deterministic_failure_modal,
                input=range(n_tasks),
                use_backups=use_backups,
                path=path,
                timing_map=timing_map,
            )
        )

    check_invocation_counts(path, timing_map, n_tasks, retries)


# fmt: off
@pytest.mark.parametrize(
    ("timing_map", "n_tasks", "retries"),
    [
        ({0: [-1], 1: [-1], 2: [-1]}, 1000, 2),
    ],
)
# fmt: on
@pytest.mark.parametrize("use_backups", [False, True])
@pytest.mark.cloud
def test_large_number_of_tasks(tmp_path, timing_map, n_tasks, retries, use_backups):
    path = f"{BASE_PATH}/{tmp_path.name}"
    outputs = asyncio.run(
        run_test(
            app_function=deterministic_failure_modal,
            input=range(n_tasks),
            use_backups=use_backups,
            path=path,
            timing_map=timing_map
        )
    )

    assert outputs == set(range(n_tasks))
    check_invocation_counts(path, timing_map, n_tasks, retries)


# fmt: off
@pytest.mark.parametrize(
    ("timing_map", "n_tasks", "retries", "expected_invocation_counts_overrides"),
    [
        # backup succeeds quickly
        ({0: [60]}, 10, 2, {0: 2}),
    ],
)
# fmt: on
@pytest.mark.cloud
def test_stragglers(tmp_path, timing_map, n_tasks, retries, expected_invocation_counts_overrides):
    path = f"{BASE_PATH}/{tmp_path.name}"
    outputs = asyncio.run(
        run_test(
            app_function=deterministic_failure_modal_long_timeout,
            input=range(n_tasks),
            path=path,
            timing_map=timing_map,
            use_backups=True,
        )
    )

    assert outputs == set(range(n_tasks))
    check_invocation_counts(path, timing_map, n_tasks, retries, expected_invocation_counts_overrides)


@pytest.mark.cloud
def test_batch(tmp_path):
    # input is unbounded, so if entire input were consumed and not read
    # in batches then it would never return, since it would never
    # run the first (failing) input
    path = f"{BASE_PATH}/{tmp_path.name}"
    with pytest.raises(RuntimeError):
        asyncio.run(
            run_test(
                app_function=deterministic_failure_modal_no_retries,
                input=itertools.count(),
                path=path,
                timing_map={0: [-1]},
                batch_size=10,
            )
        )



from cubed.primitive.memory import BufferCopies, MemoryModeller, calculate_projected_mem


def test_calculate_projected_mem():
    projected_mem = calculate_projected_mem(
        reserved_mem=3,
        inputs=[5, 7],
        operation=11,
        output=13,
        buffer_copies=BufferCopies(2, 2),
    )
    assert projected_mem == 3 + 5 + (5 * 2) + 7 + (7 * 2) + 11 + 13 + (13 * 2)


def test_memory_modeller():
    modeller = MemoryModeller()
    assert modeller.current_mem == 0
    assert modeller.peak_mem == 0

    modeller.allocate(100)
    assert modeller.current_mem == 100
    assert modeller.peak_mem == 100

    modeller.free(50)
    assert modeller.current_mem == 50
    assert modeller.peak_mem == 100



from typing import Any, Dict, Literal, Tuple, Union

T_Shape = Tuple[int, ...]

T_DType = Any  # TODO: improve this

# Regular chunks are where all chunks must be the same size, whereas
# rectangular chunks may be different sizes, and contain the size of
# each chunk along each axis.
# Zarr currently supports only regular chunks, although rectangular chunks
# are proposed in https://zarr.dev/zeps/draft/ZEP0003.html.
# Dask supports rectangular chunks.
# The general T_Chunks type is for specifying chunks, it is generally
# converted to a T_RectangularChunks or T_RegularChunks internally.
T_RegularChunks = Tuple[int, ...]
T_RectangularChunks = Tuple[Tuple[int, ...], ...]
T_Chunks = Union[
    int, T_RegularChunks, T_RectangularChunks, Dict[Any, Any], Literal["auto"]
]

T_Store = Any  # TODO: improve this



import math
import platform
import sys

import pytest

import cubed
import cubed.array_api as xp
import cubed.random
from cubed.runtime.create import create_executor
from cubed.tests.test_mem_utilization import run_operation
from cubed.tests.utils import LITHOPS_LOCAL_CONFIG

ALLOWED_MEM = 2_500_000_000

EXECUTORS = {}

if platform.system() != "Windows":
    # Run with max_tasks_per_child=1 so that each task is run in a new process,
    # allowing us to perform a stronger check on peak memory
    if sys.version_info >= (3, 11):
        executor_options = dict(max_tasks_per_child=1, max_workers=4)
        EXECUTORS["processes-single-task"] = create_executor(
            "processes", executor_options
        )

try:
    executor_options = dict(config=LITHOPS_LOCAL_CONFIG, wait_dur_sec=0.1)
    EXECUTORS["lithops"] = create_executor("lithops", executor_options)
except ImportError:
    pass


@pytest.fixture
def spec(tmp_path, reserved_mem):
    return cubed.Spec(tmp_path, allowed_mem=ALLOWED_MEM, reserved_mem=reserved_mem)


@pytest.fixture(
    scope="module",
    params=EXECUTORS.values(),
    ids=EXECUTORS.keys(),
)
def executor(request):
    return request.param


@pytest.fixture(scope="module")
def reserved_mem(executor):
    res = cubed.measure_reserved_mem(executor) * 1.1  # add some wiggle room
    return round_up_to_multiple(res, 10_000_000)  # round up to nearest multiple of 10MB


def round_up_to_multiple(x, multiple=10):
    """Round up to the nearest multiple"""
    return math.ceil(x / multiple) * multiple


@pytest.mark.slow
def test_rechunk_era5(tmp_path, spec, executor):
    # This example is based on rechunking an ERA5 dataset
    # from https://github.com/pangeo-data/rechunker/pull/89
    shape = (350640, 721, 1440)
    source_chunks = (31, 721, 1440)
    chunks = (350640, 10, 10)

    x = cubed.random.random(shape, dtype=xp.float32, chunks=source_chunks, spec=spec)

    from cubed.core.ops import _rechunk_plan

    i = 0
    for copy_chunks, target_chunks in _rechunk_plan(x, chunks):
        # Find the smallest shape that contains the three chunk sizes
        # This will be a lot less than the full ERA5 shape (350640, 721, 1440),
        # making it suitable for running in a test
        test_shape = tuple(
            max(a, b, c) for a, b, c in zip(source_chunks, copy_chunks, target_chunks)
        )
        print(i, test_shape, source_chunks, copy_chunks, target_chunks)

        a = cubed.random.random(
            test_shape, dtype=xp.float32, chunks=source_chunks, spec=spec
        )
        b = a.rechunk(target_chunks, use_new_impl=True)

        run_operation(tmp_path, executor, f"rechunk_era5_stage_{i}", b)

        source_chunks = target_chunks
        i += 1



from numbers import Integral
from typing import Any

import numpy as np
from ndindex import ndindex

from cubed.backend_array_api import namespace as nxp
from cubed.backend_array_api import numpy_array_to_backend_array
from cubed.storage.types import ArrayMetadata
from cubed.types import T_DType, T_RegularChunks, T_Shape
from cubed.utils import array_memory, broadcast_trick, memory_repr


class VirtualEmptyArray(ArrayMetadata):
    """An array that is never materialized (in memory or on disk) and contains empty values."""

    def __init__(
        self,
        shape: T_Shape,
        dtype: T_DType,
        chunks: T_RegularChunks,
    ):
        super().__init__(shape, dtype, chunks)

    def __getitem__(self, key):
        idx = ndindex[key]
        newshape = idx.newshape(self.shape)
        # use broadcast trick so array chunks only occupy a single value in memory
        return broadcast_trick(nxp.empty)(newshape, dtype=self.dtype)

    @property
    def chunkmem(self):
        # take broadcast trick into account
        return array_memory(self.dtype, (1,))


class VirtualFullArray(ArrayMetadata):
    """An array that is never materialized (in memory or on disk) and contains a single fill value."""

    def __init__(
        self,
        shape: T_Shape,
        dtype: T_DType,
        chunks: T_RegularChunks,
        fill_value: Any = None,
    ):
        super().__init__(shape, dtype, chunks)
        self.fill_value = fill_value

    def __getitem__(self, key):
        idx = ndindex[key]
        newshape = idx.newshape(self.shape)
        # use broadcast trick so array chunks only occupy a single value in memory
        return broadcast_trick(nxp.full)(
            newshape, fill_value=self.fill_value, dtype=self.dtype
        )

    @property
    def chunkmem(self):
        # take broadcast trick into account
        return array_memory(self.dtype, (1,))


class VirtualOffsetsArray(ArrayMetadata):
    """An array that is never materialized (in memory or on disk) and contains sequentially incrementing integers."""

    def __init__(self, shape: T_Shape):
        dtype = nxp.int32
        chunks = (1,) * len(shape)
        super().__init__(shape, dtype, chunks)

    def __getitem__(self, key):
        if key == () and self.shape == ():
            return nxp.asarray(0, dtype=self.dtype)
        return numpy_array_to_backend_array(
            np.ravel_multi_index(_key_to_index_tuple(key), self.shape), dtype=self.dtype
        )


class VirtualInMemoryArray(ArrayMetadata):
    """A small array that is held in memory but never materialized on disk."""

    def __init__(
        self,
        array: np.ndarray,  # TODO: generalise to array API type
        chunks: T_RegularChunks,
        max_nbytes: int = 10**6,
    ):
        nbytes = array_memory(array.dtype, array.shape)
        if nbytes > max_nbytes:
            raise ValueError(
                f"Size of in memory array is {memory_repr(nbytes)} which exceeds maximum of {memory_repr(max_nbytes)}. Consider loading the array from storage using `from_array`."
            )
        self.array = array
        super().__init__(array.shape, array.dtype, chunks)

    def __getitem__(self, key):
        return self.array.__getitem__(key)


def _key_to_index_tuple(selection):
    if isinstance(selection, slice):
        selection = (selection,)
    assert all(isinstance(s, (slice, Integral)) for s in selection)
    sel = []
    for s in selection:
        if isinstance(s, Integral):
            sel.append(s)
        elif (
            isinstance(s, slice)
            and s.stop == s.start + 1
            and (s.step is None or s.step == 1)
        ):
            sel.append(s.start)
        else:
            raise NotImplementedError(f"Offset selection not supported: {selection}")
    return tuple(sel)


def virtual_empty(
    shape: T_Shape, *, dtype: T_DType, chunks: T_RegularChunks, **kwargs
) -> VirtualEmptyArray:
    return VirtualEmptyArray(shape, dtype, chunks, **kwargs)


def virtual_full(
    shape: T_Shape,
    fill_value: Any,
    *,
    dtype: T_DType,
    chunks: T_RegularChunks,
    **kwargs,
) -> VirtualFullArray:
    return VirtualFullArray(shape, dtype, chunks, fill_value, **kwargs)


def virtual_offsets(shape: T_Shape) -> VirtualOffsetsArray:
    return VirtualOffsetsArray(shape)


def virtual_in_memory(
    array: np.ndarray,
    chunks: T_RegularChunks,
) -> VirtualInMemoryArray:
    return VirtualInMemoryArray(array, chunks)



from typing import Iterable

import numpy as np
import pytest
import zarr

import cubed
import cubed.array_api as xp
import cubed.random
from cubed._testing import assert_array_equal
from cubed.tests.utils import MAIN_EXECUTORS

icechunk = pytest.importorskip("icechunk")

from icechunk import Repository, Storage

from cubed.icechunk import store_icechunk


@pytest.fixture(
    scope="module",
    params=MAIN_EXECUTORS,
    ids=[executor.name for executor in MAIN_EXECUTORS],
)
def executor(request):
    return request.param


@pytest.fixture
def icechunk_storage(tmpdir) -> "Storage":
    return Storage.new_local_filesystem(str(tmpdir))


def create_icechunk(a, icechunk_storage, /, *, dtype=None, chunks=None):
    # from dask.asarray
    if not isinstance(getattr(a, "shape", None), Iterable):
        # ensure blocks are arrays
        a = np.asarray(a, dtype=dtype)
    if dtype is None:
        dtype = a.dtype

    repo = Repository.create(storage=icechunk_storage)
    session = repo.writable_session("main")
    store = session.store

    group = zarr.group(store=store, overwrite=True)
    arr = group.create_array("a", shape=a.shape, dtype=dtype, chunks=chunks)

    arr[...] = a

    session.commit("commit 1")


def test_from_zarr_icechunk(icechunk_storage, executor):
    create_icechunk(
        [[1, 2, 3], [4, 5, 6], [7, 8, 9]],
        icechunk_storage,
        chunks=(2, 2),
    )

    repo = Repository.open(icechunk_storage)
    session = repo.readonly_session(branch="main")
    store = session.store

    a = cubed.from_zarr(store, path="a")
    assert_array_equal(
        a.compute(executor=executor), np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    )


def test_store_icechunk(icechunk_storage, executor):
    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2))

    repo = Repository.create(storage=icechunk_storage)
    session = repo.writable_session("main")
    fork = session.fork()
    store = fork.store
    group = zarr.group(store=store, overwrite=True)
    target = group.create_array("a", shape=a.shape, dtype=a.dtype, chunks=a.chunksize)
    merged_session = store_icechunk(sources=a, targets=target, executor=executor)
    session.merge(merged_session)
    session.commit("commit 1")

    # reopen store and check contents of array
    repo = Repository.open(icechunk_storage)
    session = repo.readonly_session(branch="main")
    store = session.store

    group = zarr.open_group(store=store, mode="r")
    assert_array_equal(
        cubed.from_array(group["a"])[:], np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    )



from typing import Optional

from cubed.runtime.types import Executor


def create_executor(name: str, executor_options: Optional[dict] = None) -> Executor:
    """Create an executor from an executor name."""
    executor_options = executor_options or {}
    if name == "beam":
        from cubed.runtime.executors.beam import BeamExecutor

        return BeamExecutor(**executor_options)
    elif name == "coiled":
        from cubed.runtime.executors.coiled import CoiledExecutor

        return CoiledExecutor(**executor_options)
    elif name == "dask":
        from cubed.runtime.executors.dask import DaskExecutor

        return DaskExecutor(**executor_options)
    elif name == "lithops":
        from cubed.runtime.executors.lithops import LithopsExecutor

        return LithopsExecutor(**executor_options)
    elif name == "modal":
        from cubed.runtime.executors.modal import ModalExecutor

        return ModalExecutor(**executor_options)
    elif name == "processes":
        from cubed.runtime.executors.local import ProcessesExecutor

        return ProcessesExecutor(**executor_options)
    elif name == "ray":
        from cubed.runtime.executors.ray import RayExecutor

        return RayExecutor(**executor_options)
    elif name == "spark":
        from cubed.runtime.executors.spark import SparkExecutor

        return SparkExecutor(**executor_options)
    elif name == "single-threaded":
        from cubed.runtime.executors.local import SingleThreadedExecutor

        return SingleThreadedExecutor(**executor_options)
    elif name == "threads":
        from cubed.runtime.executors.local import ThreadsExecutor

        return ThreadsExecutor(**executor_options)
    else:
        raise ValueError(f"Unrecognized executor name: {name}")



from bisect import bisect
from operator import add, mul

import numpy as np
import tlz
from toolz import reduce

from cubed.array_api.creation_functions import empty
from cubed.backend_array_api import IS_IMMUTABLE_ARRAY
from cubed.backend_array_api import namespace as nxp
from cubed.core import (
    blockwise,
    rechunk,
    squeeze,  # noqa: F401
    unify_chunks,
)
from cubed.core.ops import (
    _create_zarr_indexer,
    elemwise,
    general_blockwise,
    map_blocks,
    map_selection,
)
from cubed.utils import (
    block_id_to_offset,
    get_item,
    normalize_chunks,
    offset_to_block_id,
    to_chunksize,
)
from cubed.vendor.dask.array.core import broadcast_chunks
from cubed.vendor.dask.array.reshape import reshape_rechunk
from cubed.vendor.dask.array.utils import validate_axis


def broadcast_arrays(*arrays):
    # From dask broadcast_arrays

    # Unify uneven chunking
    inds = [list(reversed(range(x.ndim))) for x in arrays]
    uc_args = tlz.concat(zip(arrays, inds))
    _, args = unify_chunks(*uc_args, warn=False)

    shape = np.broadcast_shapes(*(e.shape for e in args))
    chunks = broadcast_chunks(*(e.chunks for e in args))

    result = tuple(broadcast_to(e, shape=shape, chunks=chunks) for e in args)

    return result


def broadcast_to(x, /, shape, *, chunks=None):
    if x.shape == shape and (chunks is None or chunks == x.chunks):
        return x
    ndim_new = len(shape) - x.ndim
    if ndim_new < 0 or any(
        new != old for new, old in zip(shape[ndim_new:], x.shape) if old != 1
    ):
        raise ValueError(f"cannot broadcast shape {x.shape} to shape {shape}")

    if chunks is None:
        # New dimensions and broadcast dimensions have chunk size 1
        # This behaviour differs from dask where it is the full dimension size
        xchunks = normalize_chunks(x.chunks, x.shape, dtype=x.dtype)

        def chunklen(shapelen):
            return (1,) * shapelen if shapelen > 0 else (0,)

        chunks = tuple(chunklen(s) for s in shape[:ndim_new]) + tuple(
            bd if old > 1 else chunklen(new)
            for bd, old, new in zip(xchunks, x.shape, shape[ndim_new:])
        )
    else:
        chunks = normalize_chunks(
            chunks, shape, dtype=x.dtype, previous_chunks=x.chunks
        )
        for old_bd, new_bd in zip(x.chunks, chunks[ndim_new:]):
            if old_bd != new_bd and old_bd != (1,):
                raise ValueError(
                    "cannot broadcast chunks %s to chunks %s: "
                    "new chunks must either be along a new "
                    "dimension or a dimension of size 1" % (x.chunks, chunks)
                )

    # create an empty array as a template for blockwise to do broadcasting
    template = empty(shape, dtype=nxp.int8, chunks=chunks, spec=x.spec)

    return elemwise(_broadcast_like, x, template, dtype=x.dtype)


def _broadcast_like(x, template):
    return nxp.broadcast_to(x, template.shape)


def concat(arrays, /, *, axis=0, chunks=None):
    if not arrays:
        raise ValueError("Need array(s) to concat")

    if len({a.dtype for a in arrays}) > 1:
        raise ValueError("concat inputs must all have the same dtype")

    if axis is None:
        arrays = [flatten(array) for array in arrays]
        axis = 0

    if len(arrays) == 1:
        return arrays[0]

    a = arrays[0]

    # check arrays all have same shape (except in the dimension specified by axis)
    ndim = a.ndim
    if not all(
        i == axis or all(x.shape[i] == arrays[0].shape[i] for x in arrays)
        for i in range(ndim)
    ):
        raise ValueError(
            f"all the input array dimensions except for the concatenation axis must match exactly: {[x.shape for x in arrays]}"
        )

    # check arrays all have the same chunk size along axis (if more than one chunk)
    if len({a.chunksize[axis] for a in arrays if a.numblocks[axis] > 1}) > 1:
        raise ValueError(
            f"all the input array chunk sizes must match along the concatenation axis: {[x.chunksize[axis] for x in arrays]}"
        )

    # unify chunks (except in the dimension specified by axis)
    inds = [list(range(x.ndim)) for x in arrays]
    for i, ind in enumerate(inds):
        ind[axis] = -(i + 1)
    uc_args = tlz.concat(zip(arrays, inds))
    chunkss, arrays = unify_chunks(*uc_args, warn=False)

    # offsets along axis for the start of each array
    offsets = [0] + list(tlz.accumulate(add, [a.shape[axis] for a in arrays]))
    in_shapes = tuple(array.shape for array in arrays)

    axis = validate_axis(axis, ndim)
    shape = a.shape[:axis] + (offsets[-1],) + a.shape[axis + 1 :]
    dtype = a.dtype
    if chunks is None:
        # use unified chunks except for dimension specified by axis
        axis_chunksize = max(a.chunksize[axis] for a in arrays)
        chunksize = tuple(
            axis_chunksize if i == axis else chunkss[i] for i in range(ndim)
        )
        chunks = normalize_chunks(chunksize, shape=shape, dtype=dtype)
    else:
        chunks = normalize_chunks(chunks, shape=shape, dtype=dtype)

    def key_function(out_key):
        out_coords = out_key[1:]
        block_id = out_coords

        # determine the start and stop indexes for this block along the axis dimension
        chunksize = to_chunksize(chunks)
        start = block_id[axis] * chunksize[axis]
        stop = start + chunksize[axis]
        stop = min(stop, shape[axis])

        # produce a key that has slices (except for axis dimension, which is replaced below)
        idx = tuple(0 if i == axis else v for i, v in enumerate(block_id))
        key = get_item(chunks, idx)

        # find slices of the arrays
        in_keys = []
        for ai, sl in _array_slices(offsets, start, stop):
            key = tuple(sl if i == axis else k for i, k in enumerate(key))

            # use a Zarr BasicIndexer to convert this to input coordinates
            a = arrays[ai]
            indexer = _create_zarr_indexer(key, a.shape, a.chunksize)

            in_keys.extend([(a.name,) + cp.chunk_coords for cp in indexer])

        return (iter(tuple(in_key for in_key in in_keys)),)

    num_input_blocks = (1,) * len(arrays)
    iterable_input_blocks = (True,) * len(arrays)

    # We have to mark this as fusable_with_predecessors=False since the number of input args to
    # the _read_concat_chunk function is *not* the same as the number of
    # predecessor nodes in the DAG, and the fusion functions in blockwise
    # assume they are the same. See https://github.com/cubed-dev/cubed/issues/414
    # This also affects stack.
    return general_blockwise(
        _read_concat_chunk,
        key_function,
        *arrays,
        shapes=[shape],
        dtypes=[dtype],
        chunkss=[chunks],
        num_input_blocks=num_input_blocks,
        iterable_input_blocks=iterable_input_blocks,
        extra_func_kwargs=dict(dtype=dtype),
        target_shape=shape,
        target_chunks=chunks,
        axis=axis,
        offsets=offsets,
        in_shapes=in_shapes,
        function_nargs=1,
        fusable_with_predecessors=False,
    )


def _read_concat_chunk(
    arrays,
    dtype=None,
    target_shape=None,
    target_chunks=None,
    axis=None,
    offsets=None,
    in_shapes=None,
    block_id=None,
):
    # determine the start and stop indexes for this block along the axis dimension
    chunksize = to_chunksize(target_chunks)
    start = block_id[axis] * chunksize[axis]
    stop = start + chunksize[axis]
    stop = min(stop, target_shape[axis])

    chunk_shape = tuple(ch[bi] for ch, bi in zip(target_chunks, block_id))
    out = nxp.empty(chunk_shape, dtype=dtype)
    for array, (lchunk_selection, lout_selection) in zip(
        arrays,
        _chunk_slices(
            offsets, start, stop, target_chunks, chunksize, in_shapes, axis, block_id
        ),
    ):
        if IS_IMMUTABLE_ARRAY:
            out = out.at[lout_selection].set(array[lchunk_selection])
        else:
            out[lout_selection] = array[lchunk_selection]
    return out


def _array_slices(offsets, start, stop):
    """Return pairs of array index and array slice to slice from start to stop in the concatenated array."""
    slice_start = start
    while slice_start < stop:
        # find array that slice_start falls in
        i = bisect(offsets, slice_start) - 1
        slice_stop = min(stop, offsets[i + 1])
        yield i, slice(slice_start - offsets[i], slice_stop - offsets[i])
        slice_start = slice_stop


def _chunk_slices(
    offsets, start, stop, target_chunks, chunksize, in_shapes, axis, block_id
):
    """Return pairs of chunk slices to slice input array chunks and output concatenated chunk."""

    # an output chunk may have selections from more than one array, so we need an offset per array
    arr_sel_offset = 0  # offset along axis

    # produce a key that has slices (except for axis dimension, which is replaced below)
    idx = tuple(0 if i == axis else v for i, v in enumerate(block_id))
    key = get_item(target_chunks, idx)

    for ai, sl in _array_slices(offsets, start, stop):
        key = tuple(sl if i == axis else k for i, k in enumerate(key))
        indexer = _create_zarr_indexer(key, in_shapes[ai], chunksize)
        for cp in indexer:
            lout_selection_with_offset = tuple(
                sl
                if ax != axis
                else slice(sl.start + arr_sel_offset, sl.stop + arr_sel_offset)
                for ax, sl in enumerate(cp.out_selection)
            )
            yield cp.chunk_selection, lout_selection_with_offset

        arr_sel_offset += cp.out_selection[axis].stop


def expand_dims(x, /, *, axis):
    if not isinstance(axis, tuple):
        axis = (axis,)
    ndim_new = len(axis) + x.ndim
    axis = validate_axis(axis, ndim_new)

    chunks_it = iter(x.chunks)
    chunks = tuple(1 if i in axis else next(chunks_it) for i in range(ndim_new))

    return map_blocks(
        _expand_dims, x, dtype=x.dtype, chunks=chunks, new_axis=axis, axis=axis
    )


def _expand_dims(a, *args, **kwargs):
    if isinstance(a, dict):
        return {k: nxp.expand_dims(v, *args, **kwargs) for k, v in a.items()}
    return nxp.expand_dims(a, *args, **kwargs)


def flatten(x):
    return reshape(x, (-1,))


def flip(x, /, *, axis=None):
    if axis is None:
        axis = tuple(range(x.ndim))  # all axes
    if not isinstance(axis, tuple):
        axis = (axis,)
    axis = validate_axis(axis, x.ndim)

    def selection_function(out_key):
        out_coords = out_key[1:]
        block_id = out_coords

        # produce a key that has slices (except for axis dimensions, which are replaced below)
        idx = tuple(0 if i == axis else v for i, v in enumerate(block_id))
        key = list(get_item(x.chunks, idx))

        for ax in axis:
            # determine the start and stop indexes for this block along the axis dimension
            start = block_id[ax] * x.chunksize[ax]
            stop = start + x.chunksize[ax]
            stop = min(stop, x.shape[ax])

            # flip start and stop
            axis_len = x.shape[ax]
            start, stop = axis_len - stop, axis_len - start

            # replace with slice
            key[ax] = slice(start, stop)

        return tuple(key)

    max_num_input_blocks = _flip_num_input_blocks(axis, x.shape, x.chunksize)

    return map_selection(
        nxp.flip,
        selection_function,
        x,
        shape=x.shape,
        dtype=x.dtype,
        chunks=x.chunks,
        max_num_input_blocks=max_num_input_blocks,
        axis=axis,
    )


def _flip_num_input_blocks(axis, shape, chunksizes):
    num = 1
    for ax in axis:
        if shape[ax] % chunksizes[ax] != 0:
            num *= 2
    return num


def moveaxis(
    x,
    source,
    destination,
    /,
):
    # From NumPy: https://github.com/numpy/numpy/blob/a4120979d216cce00dcee511aad70bf7b45ef6e0/numpy/core/numeric.py#L1389-L1457
    from numpy.core.numeric import normalize_axis_tuple

    source = normalize_axis_tuple(source, x.ndim, "source")
    destination = normalize_axis_tuple(destination, x.ndim, "destination")
    if len(source) != len(destination):
        raise ValueError(
            "`source` and `destination` arguments must have the same number of elements"
        )

    order = [n for n in range(x.ndim) if n not in source]

    for dest, src in sorted(zip(destination, source)):
        order.insert(dest, src)

    result = permute_dims(x, order)
    return result


def permute_dims(x, /, axes):
    # From dask transpose
    if axes:
        if len(axes) != x.ndim:
            raise ValueError("axes don't match array")
    else:
        axes = tuple(range(x.ndim))[::-1]
    axes = tuple(d + x.ndim if d < 0 else d for d in axes)

    # extra memory copy due to Zarr enforcing C order on transposed array
    extra_projected_mem = x.chunkmem
    return blockwise(
        nxp.permute_dims,
        axes,
        x,
        tuple(range(x.ndim)),
        dtype=x.dtype,
        axes=axes,
        extra_projected_mem=extra_projected_mem,
    )


def repeat(x, repeats, /, *, axis=0):
    if not isinstance(repeats, int):
        raise ValueError("repeat only supports integral values for `repeats`")

    if axis is None:
        x = flatten(x)
        axis = 0

    shape = x.shape[:axis] + (x.shape[axis] * repeats,) + x.shape[axis + 1 :]
    chunks = normalize_chunks(x.chunksize, shape=shape, dtype=x.dtype)

    # This implementation calls nxp.repeat in every output block, which is 'repeats' times
    # more than necessary than if we had a primitive op that could write multiple blocks.

    def key_function(out_key):
        out_coords = out_key[1:]
        in_coords = tuple(
            bi // repeats if i == axis else bi for i, bi in enumerate(out_coords)
        )
        return ((x.name, *in_coords),)

    # extra memory from calling 'nxp.repeat' on a chunk
    extra_projected_mem = x.chunkmem * repeats
    return general_blockwise(
        _repeat,
        key_function,
        x,
        shapes=[shape],
        dtypes=[x.dtype],
        chunkss=[chunks],
        extra_projected_mem=extra_projected_mem,
        repeats=repeats,
        axis=axis,
        chunksize=x.chunksize,
    )


def _repeat(x, repeats, axis=None, chunksize=None, block_id=None):
    out = nxp.repeat(x, repeats, axis=axis)
    bi = block_id[axis] % repeats
    ind = tuple(
        slice(bi * chunksize[i], (bi + 1) * chunksize[i]) if i == axis else slice(None)
        for i in range(x.ndim)
    )
    return out[ind]


def reshape(x, /, shape, *, copy=None):
    # based on dask reshape

    known_sizes = [s for s in shape if s != -1]
    if len(known_sizes) != len(shape):
        if len(shape) - len(known_sizes) > 1:
            raise ValueError("can only specify one unknown dimension")
        # Fastpath for x.reshape(-1) on 1D arrays
        if len(shape) == 1 and x.ndim == 1:
            return x
        missing_size = x.size // reduce(mul, known_sizes, 1)
        shape = tuple(missing_size if s == -1 else s for s in shape)

    if reduce(mul, shape, 1) != x.size:
        raise ValueError("total size of new array must be unchanged")

    if x.shape == shape:
        return x

    if x.npartitions == 1:
        outchunks = tuple((d,) for d in shape)
        return reshape_chunks(x, shape, outchunks)

    inchunks, outchunks = reshape_rechunk(x.shape, shape, x.chunks)

    # TODO: make sure chunks are not too large

    x2 = rechunk(x, to_chunksize(inchunks))

    return reshape_chunks(x2, shape, outchunks)


def reshape_chunks(x, shape, chunks):
    if reduce(mul, shape, 1) != x.size:
        raise ValueError("total size of new array must be unchanged")

    # TODO: check number of chunks is unchanged
    # inchunks = normalize_chunks(x.chunks, shape=x.shape, dtype=x.dtype)
    outchunks = normalize_chunks(chunks, shape=shape, dtype=x.dtype)

    # use an empty template (handles smaller end chunks)
    template = empty(shape, dtype=x.dtype, chunks=chunks, spec=x.spec)

    def key_function(out_key):
        out_coords = out_key[1:]
        offset = block_id_to_offset(out_coords, template.numblocks)
        in_coords = offset_to_block_id(offset, x.numblocks)
        return (
            (x.name, *in_coords),
            (template.name, *out_coords),
        )

    return general_blockwise(
        _reshape_chunk,
        key_function,
        x,
        template,
        shapes=[shape],
        dtypes=[x.dtype],
        chunkss=[outchunks],
    )


def _reshape_chunk(x, template):
    return nxp.reshape(x, template.shape)


def roll(x, /, shift, *, axis=None):
    # based on dask roll
    result = x

    if axis is None:
        result = flatten(result)

        if not isinstance(shift, int):
            raise TypeError("Expect `shift` to be an int when `axis` is None.")

        shift = (shift,)
        axis = (0,)
    else:
        if not isinstance(shift, tuple):
            shift = (shift,)
        if not isinstance(axis, tuple):
            axis = (axis,)

    if len(shift) != len(axis):
        raise ValueError("Must have the same number of shifts as axes.")

    for i, s in zip(axis, shift):
        shape = result.shape[i]
        s = 0 if shape == 0 else -s % shape

        sl1 = result.ndim * [slice(None)]
        sl2 = result.ndim * [slice(None)]

        sl1[i] = slice(s, None)
        sl2[i] = slice(None, s)

        sl1 = tuple(sl1)
        sl2 = tuple(sl2)

        # note we want the concatenated array to have the same chunking as input,
        # not the chunking of result[sl1], which may be different
        result = concat([result[sl1], result[sl2]], axis=i, chunks=result.chunks)

    return reshape(result, x.shape)


def stack(arrays, /, *, axis=0):
    if not arrays:
        raise ValueError("Need array(s) to stack")

    # TODO: check arrays all have same shape
    # TODO: type promotion
    # TODO: unify chunks

    a = arrays[0]

    axis = validate_axis(axis, a.ndim + 1)
    shape = a.shape[:axis] + (len(arrays),) + a.shape[axis:]
    dtype = a.dtype
    chunks = a.chunks[:axis] + ((1,) * len(arrays),) + a.chunks[axis:]

    array_names = [a.name for a in arrays]

    def key_function(out_key):
        out_coords = out_key[1:]
        in_name = array_names[out_coords[axis]]
        return ((in_name, *(out_coords[:axis] + out_coords[(axis + 1) :])),)

    # We have to mark this as fusable_with_predecessors=False since the number of input args to
    # the _read_stack_chunk function is *not* the same as the number of
    # predecessor nodes in the DAG, and the fusion functions in blockwise
    # assume they are the same. See https://github.com/cubed-dev/cubed/issues/414
    return general_blockwise(
        _read_stack_chunk,
        key_function,
        *arrays,
        shapes=[shape],
        dtypes=[dtype],
        chunkss=[chunks],
        axis=axis,
        function_nargs=1,
        fusable_with_predecessors=False,
    )


def _read_stack_chunk(array, axis=None):
    return nxp.expand_dims(array, axis=axis)


def tile(x, repetitions, /):
    N = len(x.shape)
    M = len(repetitions)
    if N > M:
        repetitions = (1,) * (N - M) + repetitions
    elif N < M:
        for _ in range(M - N):
            x = expand_dims(x, axis=0)
    out = x
    for i, nrep in enumerate(repetitions):
        if nrep > 1:
            out = concat([out] * nrep, axis=i)
    return out


def unstack(x, /, *, axis=0):
    axis = validate_axis(axis, x.ndim)

    n_arrays = x.shape[axis]

    if n_arrays == 0:
        return ()
    elif n_arrays == 1:
        return (squeeze(x, axis=axis),)

    shape = x.shape[:axis] + x.shape[axis + 1 :]
    dtype = x.dtype
    chunks = x.chunks[:axis] + x.chunks[axis + 1 :]

    def key_function(out_key):
        out_coords = out_key[1:]
        all_in_coords = tuple(
            out_coords[:axis] + (i,) + out_coords[axis:]
            for i in range(x.numblocks[axis])
        )
        return tuple((x.name,) + in_coords for in_coords in all_in_coords)

    return general_blockwise(
        _unstack_chunk,
        key_function,
        x,
        shapes=[shape] * n_arrays,
        dtypes=[dtype] * n_arrays,
        chunkss=[chunks] * n_arrays,
        target_stores=[None] * n_arrays,  # filled in by general_blockwise
        axis=axis,
    )


def _unstack_chunk(*arrs, axis=0):
    # unstack each array in arrs and yield all in turn
    for arr in arrs:
        for a in nxp.unstack(arr, axis=axis):
            yield a



import asyncio
import copy
import time
from asyncio import Future
from typing import (
    Any,
    AsyncIterator,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    Tuple,
)

from aiostream import stream
from aiostream.core import Stream
from networkx import MultiDiGraph

from cubed.runtime.backup import should_launch_backup
from cubed.runtime.pipeline import visit_node_generations, visit_nodes
from cubed.runtime.types import Callback, CubedPipeline
from cubed.runtime.utils import (
    batched,
    handle_callbacks,
    handle_operation_start_callbacks,
)


async def async_map_unordered(
    create_futures_func: Callable[..., List[Tuple[Any, Future]]],
    input: Iterable[Any],
    use_backups: bool = False,
    create_backup_futures_func: Optional[
        Callable[..., List[Tuple[Any, Future]]]
    ] = None,
    batch_size: Optional[int] = None,
    return_stats: bool = False,
    name: Optional[str] = None,
    **kwargs,
) -> AsyncIterator[Any]:
    """
    Asynchronous parallel map over an iterable input, with support for backups and batching.
    """

    if create_backup_futures_func is None:
        create_backup_futures_func = create_futures_func

    if batch_size is None:
        inputs = input
    else:
        input_batches = batched(input, batch_size)
        inputs = next(input_batches)

    task_create_tstamp = time.time()
    tasks = {task: i for i, task in create_futures_func(inputs, name=name, **kwargs)}
    pending = set(tasks.keys())
    t = time.monotonic()
    start_times = {f: t for f in pending}
    end_times = {}
    backups: Dict[asyncio.Future, asyncio.Future] = {}

    while pending:
        finished, pending = await asyncio.wait(
            pending, return_when=asyncio.FIRST_COMPLETED, timeout=2
        )
        for task in finished:
            # TODO: use exception groups in Python 3.11 to handle case of multiple task exceptions
            if task.exception():
                # if the task has a backup that is not done, or is done with no exception, then don't raise this exception
                backup = backups.get(task, None)
                if backup:
                    if not backup.done() or not backup.exception():
                        continue
                raise task.exception()  # type: ignore
            end_times[task] = time.monotonic()
            if return_stats:
                result, stats = task.result()
                if name is not None:
                    stats["name"] = name
                stats["task_create_tstamp"] = task_create_tstamp
                yield result, stats
            else:
                yield task.result()

            # remove any backup task
            if use_backups:
                backup = backups.get(task, None)
                if backup:
                    if backup in pending:
                        pending.remove(backup)
                    del backups[task]
                    del backups[backup]
                    backup.cancel()

        if use_backups:
            now = time.monotonic()
            for task in copy.copy(pending):
                if task not in backups and should_launch_backup(
                    task, now, start_times, end_times
                ):
                    # launch backup task
                    i = tasks[task]
                    print(f"Launching backup task for input {i} at time {now}")
                    i, new_task = create_backup_futures_func([i], **kwargs)[0]
                    tasks[new_task] = i
                    start_times[new_task] = time.monotonic()
                    pending.add(new_task)
                    backups[task] = new_task
                    backups[new_task] = task

        if batch_size is not None and len(pending) < batch_size:
            inputs = next(input_batches, None)  # type: ignore
            if inputs is not None:
                new_tasks = {
                    task: i for i, task in create_futures_func(inputs, **kwargs)
                }
                tasks.update(new_tasks)
                pending.update(new_tasks.keys())
                t = time.monotonic()
                start_times = {f: t for f in new_tasks.keys()}


async def async_map_dag(
    create_futures_func: Callable,
    dag: MultiDiGraph,
    callbacks: Optional[Sequence[Callback]] = None,
    compute_arrays_in_parallel: Optional[bool] = None,
    **kwargs,
) -> None:
    """
    Asynchronous parallel map over multiple pipelines from a DAG, with support for backups and batching.
    """
    if not compute_arrays_in_parallel:
        # run one pipeline at a time
        for name, node in visit_nodes(dag):
            handle_operation_start_callbacks(callbacks, name)
            st = pipeline_to_stream(
                create_futures_func, name, node["pipeline"], **kwargs
            )
            async with st.stream() as streamer:
                async for result, stats in streamer:
                    handle_callbacks(callbacks, result, stats)
    else:
        for gen in visit_node_generations(dag):
            # run pipelines in the same topological generation in parallel by merging their streams
            streams = [
                pipeline_to_stream(
                    create_futures_func, name, node["pipeline"], **kwargs
                )
                for name, node in gen
            ]
            merged_stream = stream.merge(*streams)
            async with merged_stream.stream() as streamer:
                async for result, stats in streamer:
                    handle_callbacks(callbacks, result, stats)


def pipeline_to_stream(
    create_futures_func: Callable,
    name: str,
    pipeline: CubedPipeline,
    **kwargs,
) -> Stream:
    """
    Turn a pipeline into an asynchronous stream of results.
    """
    return stream.iterate(
        async_map_unordered(
            create_futures_func,
            pipeline.mappable,
            return_stats=True,
            name=name,
            func=pipeline.function,
            config=pipeline.config,
            **kwargs,
        )
    )



from dataclasses import asdict
from pathlib import Path
from typing import Optional

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

from cubed.runtime.types import Callback

sns.set_style("whitegrid")
matplotlib.use("Agg")


class TimelineVisualizationCallback(Callback):
    def __init__(self, format: Optional[str] = "svg") -> None:
        self.format = format

    def on_compute_start(self, event):
        self.events = []

    def on_task_end(self, event):
        self.events.append(asdict(event))

    def on_compute_end(self, event):
        events_df = pd.DataFrame(self.events)
        fig = generate_timeline(events_df)

        self.dst = Path(f"history/{event.compute_id}")
        self.dst.mkdir(parents=True, exist_ok=True)
        self.dst = self.dst / f"timeline.{self.format}"

        fig.savefig(self.dst)


def generate_timeline(events_df):
    events_df = events_df.sort_values(by=["task_create_tstamp", "name"], ascending=True)
    start_tstamp = events_df["task_create_tstamp"].min()
    total_calls = len(events_df)

    fig, ax = plt.subplots(figsize=(10, 8))

    y = np.arange(total_calls)
    point_size = 7

    fields = [
        ("task create", events_df.task_create_tstamp - start_tstamp),
        ("function start", events_df.function_start_tstamp - start_tstamp),
        ("function end", events_df.function_end_tstamp - start_tstamp),
        ("task result", events_df.task_result_tstamp - start_tstamp),
    ]

    for field_name, val in fields:
        ax.scatter(val, y, label=field_name, edgecolor="none", s=point_size, alpha=0.8)

    ax.set_xlabel("Execution time (sec)")
    ax.set_ylabel("Task number")

    ax.legend()

    return fig



import numpy as np
import pytest

from cubed._testing import assert_array_equal
from cubed.backend_array_api import namespace as nxp
from cubed.primitive.blockwise import (
    blockwise,
    general_blockwise,
    make_blockwise_key_function,
)
from cubed.runtime.executors.local import SingleThreadedExecutor
from cubed.storage.backend import open_backend_array
from cubed.tests.utils import create_zarr, execute_pipeline
from cubed.vendor.dask.blockwise import make_blockwise_graph


@pytest.fixture(scope="module", params=[SingleThreadedExecutor()])
def executor(request):
    return request.param


@pytest.mark.parametrize("reserved_mem", [0, 100])
def test_blockwise(tmp_path, executor, reserved_mem):
    source1 = create_zarr(
        [0, 1, 2], dtype=int, chunks=2, store=tmp_path / "source1.zarr"
    )
    source2 = create_zarr(
        [10, 50, 100], dtype=int, chunks=2, store=tmp_path / "source2.zarr"
    )
    allowed_mem = 1000
    target_store = tmp_path / "target.zarr"

    op = blockwise(
        nxp.linalg.outer,
        "ij",
        source1,
        "i",
        source2,
        "j",
        allowed_mem=allowed_mem,
        reserved_mem=reserved_mem,
        target_store=target_store,
        target_name="target",
        shape=(3, 3),
        dtype=int,
        chunks=(2, 2),
    )

    assert op.target_array.shape == (3, 3)
    assert op.target_array.dtype == int
    assert op.target_array.chunks == (2, 2)

    itemsize = np.dtype(int).itemsize
    assert op.projected_mem == (
        reserved_mem  # projected includes reserved
        + (itemsize * 2)  # source1 compressed chunk
        + (itemsize * 2)  # source1 uncompressed chunk
        + (itemsize * 2)  # source2 compressed chunk
        + (itemsize * 2)  # source2 uncompressed chunk
        + (itemsize * 2 * 2)  # output compressed chunk
        + (itemsize * 2 * 2)  # output uncompressed chunk
    )

    assert op.num_tasks == 4

    op.target_array.create()  # create lazy zarr array

    execute_pipeline(op.pipeline, executor=executor)

    res = open_backend_array(target_store, mode="r", path="target")
    assert_array_equal(res[:], np.outer([0, 1, 2], [10, 50, 100]))


def _permute_dims(x, /, axes, allowed_mem, reserved_mem, target_store, target_name):
    # From dask transpose
    if axes:
        if len(axes) != x.ndim:
            raise ValueError("axes don't match array")
    else:
        axes = tuple(range(x.ndim))[::-1]
    axes = tuple(d + x.ndim if d < 0 else d for d in axes)
    return blockwise(
        nxp.permute_dims,
        axes,
        x,
        tuple(range(x.ndim)),
        allowed_mem=allowed_mem,
        reserved_mem=reserved_mem,
        target_store=target_store,
        target_name=target_name,
        shape=x.shape,
        dtype=x.dtype,
        chunks=x.chunks,
        axes=axes,
    )


def test_blockwise_with_args(tmp_path, executor):
    source = create_zarr(
        [[1, 2, 3], [4, 5, 6], [7, 8, 9]],
        dtype=int,
        chunks=(2, 2),
        store=tmp_path / "source.zarr",
    )
    allowed_mem = 1000
    target_store = tmp_path / "target.zarr"

    op = _permute_dims(
        source,
        axes=(1, 0),
        allowed_mem=allowed_mem,
        reserved_mem=0,
        target_store=target_store,
        target_name="target",
    )

    assert op.target_array.shape == (3, 3)
    assert op.target_array.dtype == int
    assert op.target_array.chunks == (2, 2)

    itemsize = np.dtype(int).itemsize
    assert op.projected_mem == (
        (itemsize * 2 * 2)  # source compressed chunk
        + (itemsize * 2 * 2)  # source uncompressed chunk
        + (itemsize * 2 * 2)  # output compressed chunk
        + (itemsize * 2 * 2)  # output uncompressed chunk
    )

    assert op.num_tasks == 4

    op.target_array.create()  # create lazy zarr array

    execute_pipeline(op.pipeline, executor=executor)

    res = open_backend_array(target_store, mode="r", path="target")
    assert_array_equal(
        res[:], np.transpose(np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]), axes=(1, 0))
    )


@pytest.mark.parametrize("reserved_mem", [0, 10])
def test_blockwise_allowed_mem_exceeded(tmp_path, reserved_mem):
    source1 = create_zarr(
        [0, 1, 2], dtype=np.int64, chunks=2, store=tmp_path / "source1.zarr"
    )
    source2 = create_zarr(
        [10, 50, 100], dtype=np.int64, chunks=2, store=tmp_path / "source2.zarr"
    )
    allowed_mem = 100
    target_store = tmp_path / "target.zarr"

    with pytest.raises(
        ValueError,
        match=r"Projected blockwise memory \(\d+\) exceeds allowed_mem \(100\), including reserved_mem \(\d+\)",
    ):
        blockwise(
            nxp.linalg.outer,
            "ij",
            source1,
            "i",
            source2,
            "j",
            allowed_mem=allowed_mem,
            reserved_mem=reserved_mem,
            target_store=target_store,
            target_name="target",
            shape=(3, 3),
            dtype=np.int64,
            chunks=(2, 2),
        )


def test_general_blockwise(tmp_path, executor):
    source = create_zarr(
        list(range(20)),
        dtype=int,
        chunks=(2,),
        store=tmp_path / "source.zarr",
    )
    allowed_mem = 1000
    target_store = tmp_path / "target.zarr"

    numblocks = 10
    in_name = "x"
    merge_factor = 3

    def merge_chunks(xs):
        return nxp.concat(xs, axis=0)

    def key_function(out_key):
        out_coords = out_key[1:]

        k = merge_factor
        out_coord = out_coords[0]  # this is just 1d
        # return a tuple with a single item that is the list of input keys to be merged
        return (
            [
                (in_name, out_coord * k + i)
                for i in range(k)
                if out_coord * k + i < numblocks
            ],
        )

    op = general_blockwise(
        merge_chunks,
        key_function,
        source,
        allowed_mem=allowed_mem,
        reserved_mem=0,
        target_stores=[target_store],
        target_names=["target"],
        shapes=[(20,)],
        dtypes=[int],
        chunkss=[(6,)],
        in_names=[in_name],
    )

    assert op.target_array.shape == (20,)
    assert op.target_array.dtype == int
    assert op.target_array.chunks == (6,)

    assert op.num_tasks == 4

    op.target_array.create()  # create lazy zarr array

    execute_pipeline(op.pipeline, executor=executor)

    res = open_backend_array(target_store, mode="r", path="target")
    assert_array_equal(res[:], np.arange(20))


def test_blockwise_multiple_outputs(tmp_path, executor):
    source = create_zarr(
        [[1, 2, 3], [4, 5, 6], [7, 8, 9]],
        dtype=int,
        chunks=(2, 2),
        store=tmp_path / "source.zarr",
    )
    allowed_mem = 1000
    target_store1 = tmp_path / "target1.zarr"
    target_store2 = tmp_path / "target2.zarr"

    in_name = "x"

    def sqrts(x):
        yield np.sqrt(x)
        yield -np.sqrt(x)

    def block_function(out_key):
        out_coords = out_key[1:]
        return ((in_name, *out_coords),)

    op = general_blockwise(
        sqrts,
        block_function,
        source,
        allowed_mem=allowed_mem,
        reserved_mem=0,
        target_stores=[target_store1, target_store2],
        target_names=["target1", "target2"],
        shapes=[(3, 3), (3, 3)],
        dtypes=[float, float],
        chunkss=[(2, 2), (2, 2)],
        in_names=[in_name],
    )

    assert isinstance(op.target_array, list)
    assert len(op.target_array) == 2

    assert op.target_array[0].shape == (3, 3)
    assert op.target_array[0].dtype == float
    assert op.target_array[0].chunks == (2, 2)

    assert op.target_array[1].shape == (3, 3)
    assert op.target_array[1].dtype == float
    assert op.target_array[1].chunks == (2, 2)

    assert op.num_tasks == 4

    op.target_array[0].create()  # create lazy zarr array
    op.target_array[1].create()  # create lazy zarr array

    execute_pipeline(op.pipeline, executor=executor)

    input = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])

    res1 = open_backend_array(target_store1, mode="r", path="target1")
    assert_array_equal(res1[:], np.sqrt(input))

    res2 = open_backend_array(target_store2, mode="r", path="target2")
    assert_array_equal(res2[:], -np.sqrt(input))


def test_blockwise_multiple_outputs_fails_different_numblocks(tmp_path):
    source = create_zarr(
        [[1, 2, 3], [4, 5, 6], [7, 8, 9]],
        dtype=int,
        chunks=(2, 2),
        store=tmp_path / "source.zarr",
    )
    allowed_mem = 1000
    target_store1 = tmp_path / "target1.zarr"
    target_store2 = tmp_path / "target2.zarr"

    in_name = "x"

    def sqrts(x):
        yield np.sqrt(x)
        yield -np.sqrt(x)

    def block_function(out_key):
        out_coords = out_key[1:]
        return ((in_name, *out_coords),)

    with pytest.raises(
        ValueError,
        match="All outputs must have matching number of blocks in each dimension",
    ):
        general_blockwise(
            sqrts,
            block_function,
            source,
            allowed_mem=allowed_mem,
            reserved_mem=0,
            target_stores=[target_store1, target_store2],
            target_names=["target1", "target2"],
            shapes=[(3, 3), (3, 3)],
            dtypes=[float, float],
            chunkss=[(2, 2), (4, 2)],  # numblocks differ
            in_names=[in_name],
        )


def test_make_blockwise_key_function_map():
    func = lambda x: 0

    key_fn = make_blockwise_key_function(
        func, "z", "ij", "x", "ij", numblocks={"x": (2, 2)}
    )

    graph = make_blockwise_graph(func, "z", "ij", "x", "ij", numblocks={"x": (2, 2)})
    check_consistent_with_graph(key_fn, graph)


def test_make_blockwise_key_function_elemwise():
    func = lambda x: 0

    key_fn = make_blockwise_key_function(
        func, "z", "ij", "x", "ij", "y", "ij", numblocks={"x": (2, 2), "y": (2, 2)}
    )

    graph = make_blockwise_graph(
        func, "z", "ij", "x", "ij", "y", "ij", numblocks={"x": (2, 2), "y": (2, 2)}
    )
    check_consistent_with_graph(key_fn, graph)


def test_make_blockwise_key_function_flip():
    func = lambda x: 0

    key_fn = make_blockwise_key_function(
        func, "z", "ij", "x", "ij", "y", "ji", numblocks={"x": (2, 2), "y": (2, 2)}
    )

    graph = make_blockwise_graph(
        func, "z", "ij", "x", "ij", "y", "ji", numblocks={"x": (2, 2), "y": (2, 2)}
    )
    check_consistent_with_graph(key_fn, graph)


def test_make_blockwise_key_function_contract():
    func = lambda x: 0

    key_fn = make_blockwise_key_function(
        func, "z", "ik", "x", "ij", "y", "jk", numblocks={"x": (2, 1), "y": (1, 2)}
    )

    graph = make_blockwise_graph(
        func, "z", "ik", "x", "ij", "y", "jk", numblocks={"x": (2, 1), "y": (1, 2)}
    )
    check_consistent_with_graph(key_fn, graph)


def test_make_blockwise_key_function_contract_1d():
    func = lambda x: 0

    key_fn = make_blockwise_key_function(
        func, "z", "j", "x", "ij", numblocks={"x": (1, 2)}
    )

    graph = make_blockwise_graph(func, "z", "j", "x", "ij", numblocks={"x": (1, 2)})
    check_consistent_with_graph(key_fn, graph)


def test_make_blockwise_key_function_contract_0d():
    func = lambda x: 0

    key_fn = make_blockwise_key_function(
        func, "z", "", "x", "ij", numblocks={"x": (1, 1)}
    )

    graph = make_blockwise_graph(func, "z", "", "x", "ij", numblocks={"x": (1, 1)})
    check_consistent_with_graph(key_fn, graph)


def check_consistent_with_graph(key_fn, graph):
    for k, v in graph.items():
        assert key_fn(k) == v



import numpy as np
import pytest
import zarr

from cubed._testing import assert_array_equal
from cubed.primitive.rechunk import rechunk
from cubed.runtime.executors.local import SingleThreadedExecutor
from cubed.storage.backend import open_backend_array
from cubed.tests.utils import execute_pipeline


@pytest.fixture(scope="module", params=[SingleThreadedExecutor()])
def executor(request):
    return request.param


@pytest.mark.parametrize(
    (
        "shape",
        "source_chunks",
        "allowed_mem",
        "reserved_mem",
        "target_chunks",
        "expected_projected_mem",
        "expected_num_tasks",
    ),
    [
        # one target chunk is made up of two source chunks
        (
            (4, 4),
            (1, 2),
            1000,
            0,
            (1, 4),
            1000,
            (1,),
        ),
        # everything still works with 100 bytes of reserved_mem
        (
            (4, 4),
            (1, 2),
            1000,
            100,
            (1, 4),
            1000,
            (1,),
        ),
        # only enough memory for one source/target chunk
        (
            (4, 4),
            (1, 4),
            4 * 8 * 4,  # elts x itemsize x copies
            0,
            (4, 1),
            4 * 8 * 4,  # elts x itemsize x copies
            (16, 4),
        ),
    ],
)
def test_rechunk(
    tmp_path,
    executor,
    shape,
    source_chunks,
    allowed_mem,
    reserved_mem,
    target_chunks,
    expected_projected_mem,
    expected_num_tasks,
):
    source = zarr.ones(shape, chunks=source_chunks, store=tmp_path / "source.zarr")
    target_store = tmp_path / "target.zarr"

    ops = rechunk(
        source,
        source_array_name="source-array",
        int_array_name="int-array",
        target_array_name="target-array",
        target_chunks=target_chunks,
        allowed_mem=allowed_mem,
        reserved_mem=reserved_mem,
        target_store=target_store,
    )

    assert len(ops) == len(expected_num_tasks)

    for i, op in enumerate(ops):
        assert op.target_array.shape == shape
        assert op.target_array.dtype == source.dtype

        assert op.projected_mem == expected_projected_mem

        assert op.num_tasks == expected_num_tasks[i]

    last_op = ops[-1]
    assert last_op.target_array.chunks == target_chunks

    # create lazy zarr arrays
    for op in ops:
        op.target_array.create()

    for op in ops:
        execute_pipeline(op.pipeline, executor=executor)

    res = open_backend_array(target_store, mode="r", path="target-array")
    assert_array_equal(res[:], np.ones(shape))
    assert res.chunks == target_chunks


def test_rechunk_allowed_mem_exceeded(tmp_path):
    source = zarr.ones((4, 4), chunks=(2, 2), store=tmp_path / "source.zarr")
    allowed_mem = 16
    target_store = tmp_path / "target.zarr"

    # cubed's allowed_mem is reduced by a factor of 4 for rechunker's max_mem from 16 to 4
    with pytest.raises(
        ValueError, match=r"Source chunk memory \(32\) exceeds max_mem \(4\)"
    ):
        rechunk(
            source,
            source_array_name="source-array",
            int_array_name="int-array",
            target_array_name="target-array",
            target_chunks=(4, 1),
            allowed_mem=allowed_mem,
            reserved_mem=0,
            target_store=target_store,
        )



import zarr
import zarrs  # noqa: F401

# re-export the Zarr v3 open function
from cubed.storage.backends.zarr_python_v3 import open_zarr_v3_array  # noqa: F401

# need to set zarr config after importing from zarr_python_v3 to ensure pipeline is set
zarr.config.set(
    {
        "array.write_empty_chunks": True,
        "codec_pipeline.path": "zarrs.ZarrsCodecPipeline",
    }
)



from typing import TYPE_CHECKING, Any, List, Sequence, Union

import zarr
from icechunk.distributed import merge_sessions
from icechunk.session import ForkSession

from cubed import compute
from cubed.core.array import CoreArray
from cubed.core.ops import blockwise
from cubed.runtime.types import Callback

if TYPE_CHECKING:
    from cubed.array_api.array_object import Array


def store_icechunk(
    *,
    sources: Union["Array", Sequence["Array"]],
    targets: List[zarr.Array],
    executor=None,
    **kwargs: Any,
) -> ForkSession:
    if isinstance(sources, CoreArray):
        sources = [sources]
        targets = [targets]  # type: ignore

    if any(not isinstance(s, CoreArray) for s in sources):
        raise ValueError("All sources must be cubed array objects")

    if len(sources) != len(targets):
        raise ValueError(
            f"Different number of sources ({len(sources)}) and targets ({len(targets)})"
        )

    arrays = []
    for source, target in zip(sources, targets):
        identity = lambda a: a
        ind = tuple(range(source.ndim))
        array = blockwise(
            identity,
            ind,
            source,
            ind,
            dtype=source.dtype,
            align_arrays=False,
            target_store=target,
            return_writes_stores=True,
        )
        arrays.append(array)

    # use a callback to merge icechunk sessions
    store_callback = IcechunkStoreCallback()
    # add to other callbacks the user may have set
    callbacks = kwargs.pop("callbacks", [])
    callbacks = [store_callback] + list(callbacks)

    compute(
        *arrays,
        executor=executor,
        _return_in_memory_array=False,
        callbacks=callbacks,
        **kwargs,
    )

    return store_callback.merged_sessions


class IcechunkStoreCallback(Callback):
    def on_compute_start(self, event):
        self.sessions = []

    def on_task_end(self, event):
        result = event.result
        if result is None:
            return
        else:
            self.sessions.append(merge_sessions(*[store.session for store in result]))

    def on_compute_end(self, event):
        if len(self.sessions) == 0:
            self.merged_sessions = None
        else:
            self.merged_sessions = merge_sessions(self.sessions)



from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Dict, Optional

import memray
from memray._memray import compute_statistics
from memray._stats import Stats

from cubed.runtime.pipeline import visit_nodes
from cubed.runtime.types import Callback


class AllocationType(Enum):
    # integer values match memray AllocatorType
    MALLOC = 1
    FREE = 2
    CALLOC = 3
    REALLOC = 4
    MMAP = 10
    MUNMAP = 11


@dataclass()
class Allocation:
    object_id: str
    allocation_type: AllocationType
    memory: int
    address: Optional[int] = None
    call: Optional[str] = None

    def __repr__(self) -> str:
        return f"{self.object_id} {self.allocation_type.name} {self.memory or ''} {self.address or ''} {self.call or ''}"


class MemrayCallback(Callback):
    """Process Memray results for a computation, and print large MALLOC and FREE calls for each operation."""

    def __init__(self, mem_threshold=50_000_000) -> None:
        self.mem_threshold = mem_threshold
        self.allocations: Dict[str, Allocation] = {}
        self.stats: Dict[str, Stats] = {}

    def on_compute_end(self, event):
        for name, _ in visit_nodes(event.dag):
            memray_result_file = f"history/{event.compute_id}/memray/{name}.bin"
            if not Path(memray_result_file).is_file():
                continue

            allocations = get_allocations_over_threshold(
                memray_result_file, self.mem_threshold
            )

            print(memray_result_file)
            for allocation in allocations:
                print(allocation)

            stats = compute_statistics(memray_result_file)
            print(f"Peak memory allocated: {stats.peak_memory_allocated}")

            print()

            self.allocations[name] = allocations
            self.stats[name] = stats


def get_allocations_over_threshold(result_file, mem_threshold):
    # find all allocations over threshold and their corresponding free operations
    id = 0
    address_to_allocation = {}
    with memray.FileReader(result_file) as reader:
        for a in reader.get_allocation_records():
            if a.size >= mem_threshold:
                func, mod, line = a.stack_trace()[0]
                if a.allocator == memray.AllocatorType.MALLOC:
                    allocation_type = AllocationType.MALLOC
                elif a.allocator == memray.AllocatorType.CALLOC:
                    allocation_type = AllocationType.CALLOC
                elif a.allocator == memray.AllocatorType.REALLOC:
                    allocation_type = AllocationType.REALLOC
                elif a.allocator == memray.AllocatorType.MMAP:
                    allocation_type = AllocationType.MMAP
                else:
                    raise ValueError(f"Unsupported memray.AllocatorType {a.allocator}")
                allocation = Allocation(
                    f"object-{id:03}",
                    allocation_type,
                    a.size,
                    address=a.address,
                    call=f"{func};{mod};{line}",
                )
                id += 1
                address_to_allocation[a.address] = allocation
                yield allocation
            elif (
                a.allocator in (memray.AllocatorType.FREE, memray.AllocatorType.MUNMAP)
                and a.address in address_to_allocation
            ):
                if a.allocator == memray.AllocatorType.FREE:
                    allocation_type = AllocationType.FREE
                elif a.allocator == memray.AllocatorType.MUNMAP:
                    allocation_type = AllocationType.MUNMAP
                else:
                    raise ValueError(f"Unsupported memray.AllocatorType {a.allocator}")
                allocation = address_to_allocation.pop(a.address)
                yield Allocation(
                    allocation.object_id,
                    allocation_type,
                    allocation.memory,
                    address=a.address,
                )



import numpy as np
import pytest

import cubed
import cubed.array_api as xp
from cubed._testing import assert_allclose
from cubed.core.plan import arrays_to_plan


def test_qr():
    A = np.reshape(np.arange(32, dtype=np.float64), (16, 2))
    Q, R = xp.linalg.qr(xp.asarray(A, chunks=(4, 2)))

    plan_unopt = arrays_to_plan(Q, R)._finalize()
    assert plan_unopt.num_primitive_ops() == 4

    Q, R = cubed.compute(Q, R)

    assert_allclose(Q @ R, A, atol=1e-08)
    assert_allclose(Q.T @ Q, np.eye(2, 2), atol=1e-08)  # Q must be orthonormal
    assert_allclose(R, np.triu(R), atol=1e-08)  # R must be upper triangular


def test_qr_recursion():
    A = np.reshape(np.arange(128, dtype=np.float64), (64, 2))

    # find a memory setting where recursion happens
    found = False
    for factor in range(4, 16):
        spec = cubed.Spec(allowed_mem=128 * factor, reserved_mem=0)

        try:
            Q, R = xp.linalg.qr(xp.asarray(A, chunks=(8, 2), spec=spec))

            found = True
            plan_unopt = arrays_to_plan(Q, R)._finalize()
            assert plan_unopt.num_primitive_ops() > 4  # more than without recursion

            Q, R = cubed.compute(Q, R)

            assert_allclose(Q @ R, A, atol=1e-08)
            assert_allclose(Q.T @ Q, np.eye(2, 2), atol=1e-08)  # Q must be orthonormal
            assert_allclose(R, np.triu(R), atol=1e-08)  # R must be upper triangular

            break

        except ValueError:
            pass  # not enough memory

    assert found


def test_qr_chunking():
    A = xp.ones((32, 4), chunks=(4, 2))
    with pytest.raises(
        ValueError,
        match=r"qr only supports tall-and-skinny \(single column chunk\) arrays.",
    ):
        xp.linalg.qr(A)


def test_svd():
    A = np.reshape(np.arange(32, dtype=np.float64), (16, 2))

    U, S, Vh = xp.linalg.svd(xp.asarray(A, chunks=(4, 2)), full_matrices=False)
    U, S, Vh = cubed.compute(U, S, Vh)

    assert_allclose(U * S @ Vh, A, atol=1e-08)
    assert_allclose(U.T @ U, np.eye(2, 2), atol=1e-08)  # U must be orthonormal
    assert_allclose(Vh @ Vh.T, np.eye(2, 2), atol=1e-08)  # Vh must be orthonormal


def test_svd_recursion():
    A = np.reshape(np.arange(128, dtype=np.float64), (64, 2))

    # find a memory setting where recursion happens
    found = False
    for factor in range(4, 16):
        spec = cubed.Spec(allowed_mem=128 * factor, reserved_mem=0)

        try:
            U, S, Vh = xp.linalg.svd(
                xp.asarray(A, chunks=(8, 2), spec=spec), full_matrices=False
            )

            found = True
            plan_unopt = arrays_to_plan(U, S, Vh)._finalize()
            assert plan_unopt.num_primitive_ops() > 4  # more than without recursion

            U, S, Vh = cubed.compute(U, S, Vh)

            assert_allclose(U * S @ Vh, A, atol=1e-08)
            assert_allclose(U.T @ U, np.eye(2, 2), atol=1e-08)  # U must be orthonormal
            assert_allclose(
                Vh @ Vh.T, np.eye(2, 2), atol=1e-08
            )  # Vh must be orthonormal

            break

        except ValueError:
            pass  # not enough memory

    assert found
